"""
The wntr.network.model module includes methods to build a water network
model.
"""
import logging
from collections import OrderedDict
from typing import List, Union
from warnings import warn

import networkx as nx
import numpy as np
import pandas as pd
import six
import wntr.epanet
import wntr.network.io
from wntr.utils.ordered_set import OrderedSet

from .base import AbstractModel, Link, LinkStatus, Registry
from .controls import Control, Rule
from .elements import (
    Curve,
    Demands,
    FCValve,
    GPValve,
    HeadPump,
    Junction,
    Pattern,
    PBValve,
    Pipe,
    PowerPump,
    PRValve,
    PSValve,
    Pump,
    Reservoir,
    Source,
    Tank,
    TCValve,
    TimeSeries,
    Valve,
)

from .options import Options
from .io import read_inpfile

logger = logging.getLogger(__name__)


class WaterNetworkModel(AbstractModel):
    """
    Water network model class.

    Parameters
    -------------------
    inp_file_name: str (optional)
        Directory and filename of EPANET inp file to load into the
        :class:`~wntr.network.model.WaterNetworkModel` object.
    """

    def __init__(self, inp_file_name=None):

        # Network name
        self.name = None
        self._references: List[Union[str, dict]] = list()
        """A list of references that document the source of this model."""

        self._options = Options()
        self._node_reg = NodeRegistry(self)
        self._link_reg = LinkRegistry(self)
        self._pattern_reg = PatternRegistry(self)
        self._curve_reg = CurveRegistry(self)
        self._controls = OrderedDict()
        self._sources = SourceRegistry(self)
        self._msx = None

        self._node_reg._finalize_(self)
        self._link_reg._finalize_(self)
        self._pattern_reg._finalize_(self)
        self._curve_reg._finalize_(self)
        self._sources._finalize_(self)

        # NetworkX Graph to store the pipe connectivity and node coordinates

        self._labels = None

        self._inpfile = None
        if inp_file_name:
            # Check if internal library contains the network
            try:
                model_lib = wntr.library.model_library
                inp_file_name = model_lib.get_filepath(inp_file_name)
            except KeyError:
                pass
            read_inpfile(inp_file_name, append=self)

        # To be deleted and/or renamed and/or moved
        # Time parameters
        self.sim_time = 0.0
        self._prev_sim_time = None  # the last time at which results were accepted

    def _compare(self, other, level=1):
        """
        Parameters
        ----------
        other: WaterNetworkModel

        Returns
        -------
        bool
        """
        if (
            self.num_junctions != other.num_junctions
            or self.num_reservoirs != other.num_reservoirs
            or self.num_tanks != other.num_tanks
            or self.num_pipes != other.num_pipes
            or self.num_pumps != other.num_pumps
            or self.num_valves != other.num_valves
        ):
            return False
        for name, node in self.nodes():
            if not node._compare(other.get_node(name)):
                return False
        for name, link in self.links():
            if not link._compare(other.get_link(name)):
                return False
            
        if level > 0:
            for name, pat in self.patterns():
                if pat != other.get_pattern(name):
                    return False
            for name, curve in self.curves():
                if curve != other.get_curve(name):
                    return False
            for name, source in self.sources():
                if source != other.get_source(name):
                    return False
            if self.options != other.options:
                return False
            for name, control in self.controls():
                if not control._compare(other.get_control(name)):
                    return False
        return True

    @property
    def _shifted_time(self):
        """
        Return the time in seconds shifted by the
        simulation start time (e.g. as specified in the
        inp file). This is, this is the time since 12 AM
        on the first day.
        """
        return self.sim_time + self.options.time.start_clocktime

    @property
    def _prev_shifted_time(self):
        """
        Return the time in seconds of the previous solve shifted by
        the simulation start time. That is, this is the time from 12
        AM on the first day to the time at the previous hydraulic
        timestep.
        """
        return self._prev_sim_time + self.options.time.start_clocktime

    @property
    def _clock_time(self):
        """
        Return the current time of day in seconds from 12 AM
        """
        return self._shifted_time % (24 * 3600)

    @property
    def _clock_day(self):
        """Return the clock-time day of the simulation"""
        return int(self._shifted_time / 86400)

    ### #
    ### Iteratable attributes
    @property
    def options(self):
        """The model's options object
        
        Returns
        -------
        Options
        
        """
        return self._options

    @property
    def nodes(self):
        """The node registry (as property) or a generator for iteration (as function call)
        
        Returns
        -------
        NodeRegistry
        
        """
        return self._node_reg

    @property
    def links(self):
        """The link registry (as property) or a generator for iteration (as function call)
        
        Returns
        -------
        LinkRegistry
        
        """
        return self._link_reg

    @property
    def patterns(self):
        """The pattern registry (as property) or a generator for iteration (as function call)

        Returns
        -------
        PatternRegistry

        """
        return self._pattern_reg

    @property
    def curves(self):
        """The curve registry (as property) or a generator for iteration (as function call)
        
        Returns
        -------
        CurveRegistry        
        
        """
        return self._curve_reg

    def sources(self):
        """Returns a generator to iterate over all sources

        Returns
        -------
        A generator in the format (name, object).
        """
        for source_name, source in self._sources.items():
            yield source_name, source

    def controls(self):
        """Returns a generator to iterate over all controls

        Returns
        -------
        A generator in the format (name, object).
        """
        for control_name, control in self._controls.items():
            yield control_name, control

    ### #
    ### Element iterators
    @property
    def junctions(self):
        """Iterator over all junctions"""
        return self._node_reg.junctions

    @property
    def tanks(self):
        """Iterator over all tanks"""
        return self._node_reg.tanks

    @property
    def reservoirs(self):
        """Iterator over all reservoirs"""
        return self._node_reg.reservoirs

    @property
    def pipes(self):
        """Iterator over all pipes"""
        return self._link_reg.pipes

    @property
    def pumps(self):
        """Iterator over all pumps"""
        return self._link_reg.pumps

    @property
    def valves(self):
        """Iterator over all valves"""
        return self._link_reg.valves

    @property
    def head_pumps(self):
        """Iterator over all head-based pumps"""
        return self._link_reg.head_pumps

    @property
    def power_pumps(self):
        """Iterator over all power pumps"""
        return self._link_reg.power_pumps

    @property
    def prvs(self):
        """Iterator over all pressure reducing valves (PRVs)"""
        return self._link_reg.prvs

    @property
    def psvs(self):
        """Iterator over all pressure sustaining valves (PSVs)"""
        return self._link_reg.psvs

    @property
    def pbvs(self):
        """Iterator over all pressure breaker valves (PBVs)"""
        return self._link_reg.pbvs

    @property
    def tcvs(self):
        """Iterator over all throttle control valves (TCVs)"""
        return self._link_reg.tcvs

    @property
    def fcvs(self):
        """Iterator over all flow control valves (FCVs)"""
        return self._link_reg.fcvs

    @property
    def gpvs(self):
        """Iterator over all general purpose valves (GPVs)"""
        return self._link_reg.gpvs

    @property
    def msx(self):
        """A multispecies water quality model, if defined"""
        return self._msx
    
    @msx.setter
    def msx(self, model):
        if model is None:
            self._msx = None
            return
        from wntr.msx.base import QualityModelBase
        if not isinstance(model, QualityModelBase):
            raise TypeError('Expected QualityModelBase (or derived), got {}'.format(type(model)))
        self._msx = model

    def add_msx_model(self, msx_filename=None):
        """Add an msx model from a MSX input file (.msx extension)"""
        from wntr.msx.model import MsxModel
        self._msx = MsxModel(msx_file_name=msx_filename)

    def remove_msx_model(self):
        """Remove an msx model from the network"""
        self._msx = None

    """
    ### # 
    ### Create blank, unregistered objects (for direct assignment)
    def new_demand_timeseries_list(self):
        return Demands(self) 
    
    def new_timeseries(self):
        return TimeSeries(self, 0.0)
    
    def new_pattern(self):
        return Pattern(None, time_options=self._options.time)
    """

    ### #
    ### Add elements to the model
    def add_junction(
        self, name, base_demand=0.0, demand_pattern=None, elevation=0.0, coordinates=None, demand_category=None
    ):
        """
        Adds a junction to the water network model

        Parameters
        -------------------
        name : str
            Name of the junction.
        base_demand : float
            Base demand at the junction.
        demand_pattern : str or Pattern
            Name of the demand pattern or the Pattern object
        elevation : float
            Elevation of the junction.
        coordinates : tuple of floats
            X-Y coordinates of the node location.
        demand_category  : str
            Name of the demand category
        """
        self._node_reg.add_junction(name, base_demand, demand_pattern, elevation, coordinates, demand_category)

    def add_tank(
        self,
        name,
        elevation=0.0,
        init_level=3.048,
        min_level=0.0,
        max_level=6.096,
        diameter=15.24,
        min_vol=0.0,
        vol_curve=None,
        overflow=False,
        coordinates=None,
    ):
        """
        Adds a tank to the water network model

        Parameters
        -------------------
        name : str
            Name of the tank.
        elevation : float
            Elevation at the tank.
        init_level : float
            Initial tank level.
        min_level : float
            Minimum tank level.
        max_level : float
            Maximum tank level.
        diameter : float
            Tank diameter.
        min_vol : float
            Minimum tank volume.
        vol_curve : str, optional
            Name of a volume curve
        overflow : bool
           Overflow indicator (Always False for the WNTRSimulator)
        coordinates : tuple of float, optional
            X-Y coordinates of the node location.
            
        """
        self._node_reg.add_tank(
            name, elevation, init_level, min_level, max_level, diameter, min_vol, vol_curve, overflow, coordinates
        )

    def add_reservoir(self, name, base_head=0.0, head_pattern=None, coordinates=None):
        """
        Adds a reservoir to the water network model

        Parameters
        ----------
        name : str
            Name of the reservoir.
        base_head : float, optional
            Base head at the reservoir.
        head_pattern : str, optional
            Name of the head pattern.
        coordinates : tuple of floats, optional
            X-Y coordinates of the node location.
        
        """
        self._node_reg.add_reservoir(name, base_head, head_pattern, coordinates)

    def add_pipe(
        self,
        name,
        start_node_name,
        end_node_name,
        length=304.8,
        diameter=0.3048,
        roughness=100,
        minor_loss=0.0,
        initial_status="OPEN",
        check_valve=False,
    ):
        """
        Adds a pipe to the water network model

        Parameters
        ----------
        name : str
            Name of the pipe.
        start_node_name : str
             Name of the start node.
        end_node_name : str
             Name of the end node.
        length : float, optional
            Length of the pipe.
        diameter : float, optional
            Diameter of the pipe.
        roughness : float, optional
            Pipe roughness coefficient.
        minor_loss : float, optional
            Pipe minor loss coefficient.
        initial_status : str or LinkStatus, optional
            Pipe initial status. Options are 'OPEN' or 'CLOSED'.
        check_valve : bool, optional
            True if the pipe has a check valve.
            False if the pipe does not have a check valve.
        
        """
        self._link_reg.add_pipe(
            name, start_node_name, end_node_name, length, diameter, roughness, minor_loss, initial_status, check_valve
        )

    def add_pump(
        self,
        name,
        start_node_name,
        end_node_name,
        pump_type="POWER",
        pump_parameter=50.0,
        speed=1.0,
        pattern=None,
        initial_status="OPEN",
    ):
        """
        Adds a pump to the water network model

        Parameters
        ----------
        name : str
            Name of the pump.
        start_node_name : str
             Name of the start node.
        end_node_name : str
             Name of the end node.
        pump_type : str, optional
            Type of information provided for a pump. Options are 'POWER' or 'HEAD'.
        pump_parameter : float or str
            For a POWER pump, the pump power.
            For a HEAD pump, the head curve name.
        speed: float
            Relative speed setting (1.0 is normal speed)
        pattern: str
            Name of the speed pattern
        initial_status : str or LinkStatus
            Pump initial status. Options are 'OPEN' or 'CLOSED'.
        
        """
        self._link_reg.add_pump(
            name, start_node_name, end_node_name, pump_type, pump_parameter, speed, pattern, initial_status
        )

    def add_valve(
        self,
        name,
        start_node_name,
        end_node_name,
        diameter=0.3048,
        valve_type="PRV",
        minor_loss=0.0,
        initial_setting=0.0,
        initial_status="ACTIVE",
    ):
        """
        Adds a valve to the water network model

        Parameters
        ----------
        name : str
            Name of the valve.
        start_node_name : str
             Name of the start node.
        end_node_name : str
             Name of the end node.
        diameter : float, optional
            Diameter of the valve.
        valve_type : str, optional
            Type of valve. Options are 'PRV', 'PSV', 'PBV', 'FCV', 'TCV', and 'GPV'
        minor_loss : float, optional
            Pipe minor loss coefficient.
        initial_setting : float or str, optional
            Valve initial setting.
            Pressure setting for PRV, PSV, or PBV. 
            Flow setting for FCV. 
            Loss coefficient for TCV.
            Name of headloss curve for GPV.
        initial_status: str or LinkStatus
            Valve initial status. Options are 'OPEN',  'CLOSED', or 'ACTIVE'.
        """
        self._link_reg.add_valve(
            name, start_node_name, end_node_name, diameter, valve_type, minor_loss, initial_setting, initial_status
        )

    def add_pattern(self, name, pattern=None):
        """
        Adds a pattern to the water network model
        
        The pattern can be either a list of values (list, numpy array, etc.) or a 
        :class:`~wntr.network.elements.Pattern` object. The Pattern class has options to automatically
        create certain types of patterns, such as a single, on/off pattern (previously created using
        the start_time and stop_time arguments to this function) -- see the class documentation for
        examples.

        
        .. warning::
            Patterns **must** be added to the model prior to adding any model element that uses the pattern,
            such as junction demands, sources, etc. Patterns are linked by reference, so changes to a 
            pattern affects all elements using that pattern. 

            
        .. warning::
            Patterns **always** use the global water network model options.time values.
            Patterns **will not** be resampled to match these values, it is assumed that 
            patterns created using Pattern(...) or Pattern.binary_pattern(...) object used the same 
            pattern timestep value as the global value, and they will be treated accordingly.


        Parameters
        ----------
        name : str
            Name of the pattern.
        pattern : list of floats or Pattern
            A list of floats that make up the pattern, or a :class:`~wntr.network.elements.Pattern` object.

        Raises
        ------
        ValueError
            If adding a pattern with `name` that already exists.

        
        """
        self._pattern_reg.add_pattern(name, pattern)

    def add_curve(self, name, curve_type, xy_tuples_list):
        """
        Adds a curve to the water network model

        Parameters
        ----------
        name : str
            Name of the curve.
        curve_type : str
            Type of curve. Options are HEAD, EFFICIENCY, VOLUME, HEADLOSS.
        xy_tuples_list : list of (x, y) tuple
            List of X-Y coordinate tuples on the curve.
        """
        self._curve_reg.add_curve(name, curve_type, xy_tuples_list)

    def add_source(self, name, node_name, source_type, quality, pattern=None):
        """
        Adds a source to the water network model

        Parameters
        ----------
        name : str
            Name of the source

        node_name: str
            Injection node.

        source_type: str
            Source type, options = CONCEN, MASS, FLOWPACED, or SETPOINT

        quality: float
            Source strength in Mass/Time for MASS and Mass/Volume for CONCEN, 
            FLOWPACED, or SETPOINT

        pattern: str or Pattern
            Pattern name or object
        """
        if pattern and isinstance(pattern, six.string_types):
            pattern = self.get_pattern(pattern)
        source = Source(self, name, node_name, source_type, quality, pattern)
        self._sources[source.name] = source
        self._pattern_reg.add_usage(source.strength_timeseries.pattern_name, (source.name, "Source"))
        self._node_reg.add_usage(source.node_name, (source.name, "Source"))

    def add_control(self, name, control_object):
        """
        Adds a control or rule to the water network model

        Parameters
        ----------
        name : str
           control object name.
        control_object : Control/Rule object or string
            Control object, Rule object, or string with the control. The string
            must be written in EPANET INP file format and SI units 
            (i.e., "LINK Pump-1 OPEN IF NODE Tank-1 BELOW 10")
        """
        if isinstance(control_object, str):
            line = control_object
            flow_units = wntr.epanet.util.FlowUnits['SI']
            control_object = wntr.epanet.io._read_control_line(line, self, flow_units, name)
        
        if name in self._controls:
            raise ValueError(
                "The name provided for the control is already used. Please either remove the control with that name first or use a different name for this control."
            )
        self._controls[name] = control_object

    ### #
    ### Remove elements from the model
    def remove_node(self, name, with_control=False, force=False):
        """Removes a node from the water network model"""
        node = self.get_node(name)
        if not force:
            if with_control:
                x = []
                for control_name, control in self._controls.items():
                    if node in control.requires():
                        logger.warning(
                            control._control_type_str()
                            + " "
                            + control_name
                            + " is being removed along with node "
                            + name
                        )
                        x.append(control_name)
                for i in x:
                    self.remove_control(i)
            else:
                for control_name, control in self._controls.items():
                    if node in control.requires():
                        raise RuntimeError(
                            "Cannot remove node {0} without first removing control/rule {1}".format(name, control_name)
                        )
        self._node_reg.__delitem__(name)

    def remove_link(self, name, with_control=False, force=False):
        """Removes a link from the water network model"""
        link = self.get_link(name)
        if not force:
            if with_control:
                x = []
                for control_name, control in self._controls.items():
                    if link in control.requires():
                        logger.warning(
                            control._control_type_str()
                            + " "
                            + control_name
                            + " is being removed along with link "
                            + name
                        )
                        x.append(control_name)
                for i in x:
                    self.remove_control(i)
            else:
                for control_name, control in self._controls.items():
                    if link in control.requires():
                        raise RuntimeError(
                            "Cannot remove link {0} without first removing control/rule {1}".format(name, control_name)
                        )
        self._link_reg.__delitem__(name)

    def remove_pattern(self, name):
        """Removes a pattern from the water network model"""
        self._pattern_reg.__delitem__(name)

    def remove_curve(self, name):
        """Removes a curve from the water network model"""
        self._curve_reg.__delitem__(name)

    def remove_source(self, name):
        """Removes a source from the water network model

        Parameters
        ----------
        name : str
           The name of the source object to be removed
        """
        logger.warning(
            "You are deleting a source. This could have unintended \
            side effects. If you are replacing values, use get_source(name) \
            and modify it instead."
        )
        source = self._sources[name]
        self._pattern_reg.remove_usage(source.strength_timeseries.pattern_name, (source.name, "Source"))
        self._node_reg.remove_usage(source.node_name, (source.name, "Source"))
        del self._sources[name]

    def remove_control(self, name):
        """Removes a control from the water network model"""
        del self._controls[name]

    def _discard_control(self, name):
        """Removes a control from the water network model
        
        If the control is not present, an exception is not raised.

        Parameters
        ----------
        name : string
           The name of the control object to be removed.
        """
        try:
            del self._controls[name]
        except KeyError:
            pass

    ### #
    ### Get elements from the model
    def get_node(self, name):
        """Get a specific node
        
        Parameters
        ----------
        name : str
            The node name
            
        Returns
        -------
        Junction, Tank, or Reservoir
        
        """
        return self._node_reg[name]

    def get_link(self, name):
        """Get a specific link
        
        Parameters
        ----------
        name : str
            The link name
            
        Returns
        -------
        Pipe, Pump, or Valve
        
        """
        return self._link_reg[name]

    def get_pattern(self, name):
        """Get a specific pattern
        
        Parameters
        ----------
        name : str
            The pattern name
            
        Returns
        -------
        Pattern
        
        """
        return self._pattern_reg[name]

    def get_curve(self, name):
        """Get a specific curve
        
        Parameters
        ----------
        name : str
            The curve name
            
        Returns
        -------
        Curve
        
        """
        return self._curve_reg[name]

    def get_source(self, name):
        """Get a specific source
        
        Parameters
        ----------
        name : str
            The source name
            
        Returns
        -------
        Source
        
        """
        return self._sources[name]

    def get_control(self, name):
        """Get a specific control or rule
        
        Parameters
        ----------
        name : str
            The control name
            
        Returns
        -------
        ctrl: Control or Rule
        
        """
        return self._controls[name]

    ### #
    ### Name lists
    @property
    def node_name_list(self):
        """Get a list of node names
        
        Returns
        -------
        list of strings
        
        """
        return list(self._node_reg.keys())

    @property
    def junction_name_list(self):
        """Get a list of junction names
        
        Returns
        -------
        list of strings
        
        """
        return list(self._node_reg.junction_names)

    @property
    def tank_name_list(self):
        """Get a list of tanks names
        
        Returns
        -------
        list of strings
        
        """
        return list(self._node_reg.tank_names)

    @property
    def reservoir_name_list(self):
        """Get a list of reservoir names
        
        Returns
        -------
        list of strings
        
        """
        return list(self._node_reg.reservoir_names)

    @property
    def link_name_list(self):
        """Get a list of link names
        
        Returns
        -------
        list of strings
        
        """
        return list(self._link_reg.keys())

    @property
    def pipe_name_list(self):
        """Get a list of pipe names
        
        Returns
        -------
        list of strings
        
        """
        return list(self._link_reg.pipe_names)

    @property
    def pump_name_list(self):
        """Get a list of pump names (both types included)

        Returns
        -------
        list of strings

        """
        return list(self._link_reg.pump_names)

    @property
    def head_pump_name_list(self):
        """Get a list of head pump names

        Returns
        -------
        list of strings

        """
        return list(self._link_reg.head_pump_names)

    @property
    def power_pump_name_list(self):
        """Get a list of power pump names

        Returns
        -------
        list of strings

        """
        return list(self._link_reg.power_pump_names)

    @property
    def valve_name_list(self):
        """Get a list of valve names (all types included)

        Returns
        -------
        list of strings

        """
        return list(self._link_reg.valve_names)

    @property
    def prv_name_list(self):
        """Get a list of prv names

        Returns
        -------
        list of strings

        """
        return list(self._link_reg.prv_names)

    @property
    def psv_name_list(self):
        """Get a list of psv names

        Returns
        -------
        list of strings

        """
        return list(self._link_reg.psv_names)

    @property
    def pbv_name_list(self):
        """Get a list of pbv names

        Returns
        -------
        list of strings

        """
        return list(self._link_reg.pbv_names)

    @property
    def tcv_name_list(self):
        """Get a list of tcv names

        Returns
        -------
        list of strings

        """
        return list(self._link_reg.tcv_names)

    @property
    def fcv_name_list(self):
        """Get a list of fcv names

        Returns
        -------
        list of strings

        """
        return list(self._link_reg.fcv_names)

    @property
    def gpv_name_list(self):
        """Get a list of gpv names

        Returns
        -------
        list of strings

        """
        return list(self._link_reg.gpv_names)

    @property
    def pattern_name_list(self):
        """Get a list of pattern names
        
        Returns
        -------
        list of strings
        
        """
        return list(self._pattern_reg.keys())

    @property
    def curve_name_list(self):
        """Get a list of curve names
        
        Returns
        -------
        list of strings
        
        """
        return list(self._curve_reg.keys())

    @property
    def source_name_list(self):
        """Get a list of source names
        
        Returns
        -------
        list of strings
        
        """
        return list(self._sources.keys())

    @property
    def control_name_list(self):
        """Get a list of control/rule names
        
        Returns
        -------
        list of strings
        
        """
        return list(self._controls.keys())

    ### #
    ### Counts
    @property
    def num_nodes(self):
        """The number of nodes"""
        return len(self._node_reg)

    @property
    def num_junctions(self):
        """The number of junctions"""
        return len(self._node_reg.junction_names)

    @property
    def num_tanks(self):
        """The number of tanks"""
        return len(self._node_reg.tank_names)

    @property
    def num_reservoirs(self):
        """The number of reservoirs"""
        return len(self._node_reg.reservoir_names)

    @property
    def num_links(self):
        """The number of links"""
        return len(self._link_reg)

    @property
    def num_pipes(self):
        """The number of pipes"""
        return len(self._link_reg.pipe_names)

    @property
    def num_pumps(self):
        """The number of pumps"""
        return len(self._link_reg.pump_names)

    @property
    def num_valves(self):
        """The number of valves"""
        return len(self._link_reg.valve_names)

    @property
    def num_patterns(self):
        """The number of patterns"""
        return len(self._pattern_reg)

    @property
    def num_curves(self):
        """The number of curves"""
        return len(self._curve_reg)

    @property
    def num_sources(self):
        """The number of sources"""
        return len(self._sources)

    @property
    def num_controls(self):
        """The number of controls"""
        return len(self._controls)

    ### #
    ### Helper functions
    def describe(self, level=0):
        """
        Describe number of components in the network model
        
        Parameters
        ----------
        level : int (0, 1, or 2)
            
           * Level 0 returns the number of Nodes, Links, Patterns, Curves, Sources, and Controls.
           * Level 1 includes information from Level 0 but 
             divides Nodes into Junctions, Tanks, and Reservoirs, 
             divides Links into Pipes, Pumps, and Valves, and 
             divides Curves into Pump, Efficiency, Headloss, and Volume.
           * Level 2 includes information from Level 1 but 
             divides Pumps into Head and Power, and 
             divides Valves into PRV, PSV, PBV, TCV, FCV, and GPV.
            
        Returns
        -------
        A dictionary with component counts
        """

        d = {
            "Nodes": self.num_nodes,
            "Links": self.num_links,
            "Patterns": self.num_patterns,
            "Curves": self.num_curves,
            "Sources": self.num_sources,
            "Controls": self.num_controls,
        }

        if level >= 1:
            d["Nodes"] = {"Junctions": self.num_junctions, "Tanks": self.num_tanks, "Reservoirs": self.num_reservoirs}
            d["Links"] = {"Pipes": self.num_pipes, "Pumps": self.num_pumps, "Valves": self.num_valves}
            d["Curves"] = {
                "Pump": len(self._curve_reg._pump_curves),
                "Efficiency": len(self._curve_reg._efficiency_curves),
                "Headloss": len(self._curve_reg._headloss_curves),
                "Volume": len(self._curve_reg._volume_curves),
            }

        if level >= 2:
            d["Links"]["Pumps"] = {"Head": len(list(self.head_pumps())), "Power": len(list(self.power_pumps()))}
            d["Links"]["Valves"] = {
                "PRV": len(list(self.prvs())),
                "PSV": len(list(self.psvs())),
                "PBV": len(list(self.pbvs())),
                "TCV": len(list(self.tcvs())),
                "FCV": len(list(self.fcvs())),
                "GPV": len(list(self.gpvs())),
            }

        return d

    def to_dict(self):
        """
        Dictionary representation of the WaterNetworkModel.
        
        Returns
        -------
        dict
        """
        return wntr.network.io.to_dict(self)

    def from_dict(self, d: dict):
        """
        Append the model with elements from a dictionary.

        Parameters
        ----------
        d : dict
            Dictionary representation of the WaterNetworkModel
        """
        wntr.network.io.from_dict(d, append=self)

    def to_gis(self, crs=None, pumps_as_points=False, valves_as_points=False):
        """
        Convert a WaterNetworkModel into GeoDataFrames
        
        Parameters
        ----------
        crs : str, optional
            Coordinate reference system, by default None
        pumps_as_points : bool, optional
            Represent pumps as points (True) or lines (False), by default False
        valves_as_points : bool, optional
            Represent valves as points (True) or lines (False), by default False
            
        Returns
        -------
        WaterNetworkGIS object that contains junctions, tanks, reservoirs, pipes, 
        pumps, and valves GeoDataFrames
        """
        return wntr.network.io.to_gis(self, crs, pumps_as_points, valves_as_points)
    
    def from_gis(self, gis_data):
        """
        Append the model with elements from GeoDataFrames
        
        Parameters
        ----------
        gis_data : WaterNetworkGIS or dict of geopandas.GeoDataFrame
            GeoDataFrames containing water network attributes. If gis_data is a 
            dictionary, then the keys are junctions, tanks, reservoirs, pipes, 
            pumps, and valves. If the pumps or valves are Points, they will be 
            converted to Lines with the same start and end node location.
            
        Returns
        -------
        WaterNetworkModel
        """
        return wntr.network.io.from_gis(gis_data, append=self)
    
    def to_graph(self, node_weight=None, link_weight=None, 
                 modify_direction=False):
        """
        Convert a WaterNetworkModel into a networkx MultiDiGraph
        
        Parameters
        ----------
        node_weight :  dict or pandas Series (optional)
            Node weights
        link_weight : dict or pandas Series (optional)
            Link weights.  
        modify_direction : bool (optional)
            If True, than if the link weight is negative, the link start and 
            end node are switched and the abs(weight) is assigned to the link
            (this is useful when weighting graphs by flowrate). If False, link 
            direction and weight are not changed.
            
        Returns
        --------
        networkx MultiDiGraph
        """
        return wntr.network.io.to_graph(self, node_weight, link_weight, 
                                        modify_direction)
                                        
                               
    def get_graph(self, node_weight=None, link_weight=None, modify_direction=False):
        """
        Convert a :class:`~wntr.network.model.WaterNetworkModel` into a networkx MultiDiGraph
        
        .. deprecated:: 0.5.0
        Use :meth:`~wntr.network.model.WaterNetworkModel.to_graph` instead
        
        Parameters
        ----------
        node_weight :  dict or pandas.Series (optional)
            Node weights
        link_weight : dict or pandas.Series (optional)
            Link weights.  
        modify_direction : bool (optional)
            If True, then if the link weight is negative, the link start and 
            end node are switched and the abs(weight) is assigned to the link
            (this is useful when weighting graphs by flowrate). If False, link 
            direction and weight are not changed.
            
        Returns
        --------
        networkx MultiDiGraph
        """
        warn("wntr.network.WaterNetworkModel.get_graph is deprecated, use wntr.network.WaterNetworkModel.to_graph instead", DeprecationWarning, stacklevel=2)
        
        return wntr.network.io.to_graph(self, node_weight, link_weight, 
                                        modify_direction)

    def assign_demand(self, demand, pattern_prefix="ResetDemand"):
        """
        Assign demands using values in a DataFrame. 
        
        New demands are specified in a pandas DataFrame indexed by
        time (in seconds). The method resets junction demands by creating a 
        new demand pattern and using a base demand of 1. 
        The demand pattern is resampled to match the water network model 
        pattern timestep. This method can be
        used to reset demands in a water network model to demands from a
        pressure dependent demand simulation.

        Parameters
        ----------
        demand : pandas.DataFrame
            A pandas DataFrame containing demands (index = time, columns = junction names)

        pattern_prefix: str
            Pattern name prefix, default = 'ResetDemand'.  The junction name is 
            appended to the prefix to create a new pattern name.  
            If the pattern name already exists, an error is thrown and the user 
            should use a different pattern prefix name.
        """
        for junc_name in demand.columns:

            # Extract the node demand pattern and resample to match the pattern timestep
            demand_pattern = demand.loc[:, junc_name]
            demand_pattern.index = pd.to_timedelta(demand_pattern.index, "s")
            resample_offset = str(int(self.options.time.pattern_timestep)) + "s"
            demand_pattern = demand_pattern.resample(resample_offset).mean() / self.options.hydraulic.demand_multiplier

            # Add the pattern
            # If the pattern name already exists, this fails
            pattern_name = pattern_prefix + junc_name
            self.add_pattern(pattern_name, demand_pattern.tolist())

            # Reset base demand
            junction = self.get_node(junc_name)
            junction.demand_timeseries_list.clear()
            junction.demand_timeseries_list.append((1.0, pattern_name))

    def get_links_for_node(self, node_name, flag="ALL"):
        """
        Returns a list of links connected to a node

        Parameters
        ----------
        node_name : str
            Name of the node.

        flag : str
            Options are 'ALL', 'INLET', 'OUTLET'.
            'ALL' returns all links connected to the node.
            'INLET' returns links that have the specified node as an end node.
            'OUTLET' returns links that have the specified node as a start node.

        Returns
        -------
        A list of link names connected to the node
        """
        link_types = {"Pipe", "Pump", "Valve"}
        link_data = self._node_reg.get_usage(node_name)
        if link_data is None:
            return []
        else:
            if flag.upper() == "ALL":
                return [
                    link_name
                    for link_name, link_type in link_data
                    if link_type in link_types
                    and node_name in {self.get_link(link_name).start_node_name, self.get_link(link_name).end_node_name}
                ]
            elif flag.upper() == "INLET":
                return [
                    link_name
                    for link_name, link_type in link_data
                    if link_type in link_types and node_name == self.get_link(link_name).end_node_name
                ]
            elif flag.upper() == "OUTLET":
                return [
                    link_name
                    for link_name, link_type in link_data
                    if link_type in link_types and node_name == self.get_link(link_name).start_node_name
                ]
            else:
                logger.error("Unrecognized flag: {0}".format(flag))
                raise ValueError("Unrecognized flag: {0}".format(flag))

    def query_node_attribute(self, attribute, operation=None, value=None, node_type=None):
        """
        Query node attributes, for example get all nodes with elevation <= threshold

        Parameters
        ----------
        attribute: str
            :class:`~wntr.network.base.Node` attribute.

        operation: numpy operator
            Numpy operator, options include
            :obj:`numpy.greater`,
            :obj:`numpy.greater_equal`,
            :obj:`numpy.less`,
            :obj:`numpy.less_equal`,
            :obj:`numpy.equal`,
            :obj:`numpy.not_equal`.

        value: float or int
            Threshold

        node_type: Node type
            :class:`~wntr.network.base.Node` type, options include
            :class:`~wntr.network.base.Node`,
            :class:`~wntr.network.elements.Junction`,
            :class:`~wntr.network.elements.Reservoir`,
            :class:`~wntr.network.elements.Tank`, or None. Default = None.
            Note None and :class:`~wntr.network.base.Node` produce the same results.

        Returns
        -------
        :class:`pandas.Series` that contains the attribute that satisfies the operation threshold for a given node_type.

        Notes
        -----
        If operation and value are both None, the :class:`pandas.Series` will contain the attributes
        for all nodes with the specified attribute.

        """
        node_attribute_dict = {}
        for name, node in self.nodes(node_type):
            try:
                if operation == None and value == None:
                    node_attribute_dict[name] = getattr(node, attribute)
                else:
                    node_attribute = getattr(node, attribute)
                    if operation(node_attribute, value):
                        node_attribute_dict[name] = node_attribute
            except AttributeError:
                pass
        return pd.Series(node_attribute_dict)

    def query_link_attribute(self, attribute, operation=None, value=None, link_type=None):
        """
        Query link attributes, for example get all pipe diameters > threshold

        Parameters
        ----------
        attribute: str
            :class:`~wntr.network.base.Link` attribute

        operation: numpy operator
            Numpy operator, options include the following: 
            :obj:`numpy.greater`,
            :obj:`numpy.greater_equal`,
            :obj:`numpy.less`,
            :obj:`numpy.less_equal`,
            :obj:`numpy.equal`,
            :obj:`numpy.not_equal`.

        value: float or int
            Threshold

        link_type: Link type
            :class:`~wntr.network.base.Link` type, options include
            :class:`~wntr.network.base.Link`,
            :class:`~wntr.network.elements.Pipe`,
            :class:`~wntr.network.elements.Pump`,
            :class:`~wntr.network.elements.Valve`, or None. Default = None.
            Note None and :class:`~wntr.network.base.Link` produce the same results.
    
        Returns
        -------
        :class:`pandas.Series` that contains the attribute that satisfies the operation threshold for a given link_type.

        Notes
        -----
        If operation and value are both None, the :class:`pandas.Series` will contain the attributes
        for all links with the specified attribute.

        """
        link_attribute_dict = {}
        for name, link in self.links(link_type):
            try:
                if operation == None and value == None:
                    link_attribute_dict[name] = getattr(link, attribute)
                else:
                    link_attribute = getattr(link, attribute)
                    if operation(link_attribute, value):
                        link_attribute_dict[name] = link_attribute
            except AttributeError:
                pass
        return pd.Series(link_attribute_dict)

    def convert_controls_to_rules(self, priority=3):
        """
        Convert all controls to rules.
        
        Note that for an exact match between controls and rules, the rule 
        timestep must be very small.

        Parameters
        ----------
        priority : int, optional
           Rule priority, default is 3.

        """
        for name in self.control_name_list:
            control = self.get_control(name)
            if isinstance(control, Control):
                act = control.actions()[0]
                cond = control.condition
                rule = Rule(cond, act, priority=priority)
                self.add_control(name.replace(" ", "_") + "_Rule", rule)
                self.remove_control(name)

    def reset_initial_values(self):
        """
        Resets all initial values in the network
        """
        #### TODO: move reset conditions to /sim
        self.sim_time = 0.0
        self._prev_sim_time = None

        for name, node in self.nodes(Junction):
            node._head = None
            node._demand = None
            node._pressure = None
            node._leak_demand = None
            node._leak_status = False
            node._is_isolated = False

        for name, node in self.nodes(Tank):
            node._head = node.init_level + node.elevation
            node._prev_head = node.head
            node._demand = None
            node._leak_demand = None
            node._leak_status = False
            node._is_isolated = False

        for name, node in self.nodes(Reservoir):
            node._head = None  # node.head_timeseries.base_value
            node._demand = None
            node._leak_demand = None
            node._is_isolated = False

        for name, link in self.links(Pipe):
            link._user_status = link.initial_status
            link._setting = link.initial_setting
            link._internal_status = LinkStatus.Active
            link._is_isolated = False
            link._flow = None
            link._prev_setting = None

        for name, link in self.links(Pump):
            link._user_status = link.initial_status
            link._setting = link.initial_setting
            link._internal_status = LinkStatus.Active
            link._is_isolated = False
            link._flow = None
            if isinstance(link, PowerPump):
                link.power = link._base_power
            link._prev_setting = None

        for name, link in self.links(Valve):
            link._user_status = link.initial_status
            link._setting = link.initial_setting
            link._internal_status = LinkStatus.Active
            link._is_isolated = False
            link._flow = None
            link._prev_setting = None

        for name, control in self.controls():
            control._reset()


class PatternRegistry(Registry):
    """A registry for patterns."""

    def _finalize_(self, model):
        super()._finalize_(model)
        self._pattern_reg = None

    class DefaultPattern(object):
        """An object that always points to the current default pattern for a model"""

        def __init__(self, options):
            self._options = options

        def __str__(self):
            return str(self._options.hydraulic.pattern) if self._options.hydraulic.pattern is not None else ""

        def __repr__(self):
            return "DefaultPattern()"

        @property
        def name(self):
            """The name of the default pattern, or ``None`` if no pattern is assigned"""
            return str(self._options.hydraulic.pattern) if self._options.hydraulic.pattern is not None else ""

    def __getitem__(self, key):
        try:
            return super(PatternRegistry, self).__getitem__(key)
        except KeyError:
            return None

    def add_pattern(self, name, pattern=None):
        """
        Adds a pattern to the water network model.
        
        The pattern can be either a list of values (list, numpy array, etc.) or 
        a :class:`~wntr.network.elements.Pattern` object. The Pattern class has 
        options to automatically create certain types of patterns, such as a 
        single, on/off pattern

        .. warning::
            Patterns **must** be added to the model prior to adding any model 
            element that uses the pattern, such as junction demands, sources, 
            etc. Patterns are linked by reference, so changes to a pattern 
            affects all elements using that pattern. 

        .. warning::
            Patterns **always** use the global water network model options.time 
            values. Patterns **will not** be resampled to match these values, 
            it is assumed that patterns created using Pattern(...) or 
            Pattern.binary_pattern(...) object used the same pattern timestep 
            value as the global value, and they will be treated accordingly.

        Parameters
        ----------
        name : str
            Name of the pattern.
        pattern : list of float or Pattern
            A list of floats that make up the pattern, or a 
            :class:`~wntr.network.elements.Pattern` object.

        Raises
        ------
        ValueError
            If adding a pattern with `name` that already exists.
        """
        assert (
            isinstance(name, str) and len(name) < 32 and name.find(" ") == -1
        ), "name must be a string with less than 32 characters and contain no spaces"
        assert isinstance(pattern, (list, np.ndarray, Pattern)), "pattern must be a list or Pattern"

        if not isinstance(pattern, Pattern):
            pattern = Pattern(name, multipliers=pattern, time_options=self._options.time)
        else:  # elif pattern.time_options is None:
            pattern.time_options = self._options.time
        if pattern.name in self._data.keys():
            raise ValueError("Pattern name already exists")
        self[name] = pattern

    @property
    def default_pattern(self):
        """A new default pattern object"""
        return self.DefaultPattern(self._options)


class CurveRegistry(Registry):
    """A registry for curves."""

    def __init__(self, model):
        super(CurveRegistry, self).__init__(model)
        self._pump_curves = OrderedSet()
        self._efficiency_curves = OrderedSet()
        self._headloss_curves = OrderedSet()
        self._volume_curves = OrderedSet()

    def _finalize_(self, model):
        super()._finalize_(model)
        self._curve_reg = None

    def __setitem__(self, key, value):
        if not isinstance(key, six.string_types):
            raise ValueError("Registry keys must be strings")
        self._data[key] = value
        if value is not None:
            self.set_curve_type(key, value.curve_type)

    def set_curve_type(self, key, curve_type):
        """
        Sets curve type.
        
        WARNING -- this does not check to make sure key is typed before assigning it - 
        you could end up with a curve that is used for more than one type"""
        if curve_type is None:
            return
        curve_type = curve_type.upper()
        if curve_type == "HEAD":
            self._pump_curves.add(key)
        elif curve_type == "HEADLOSS":
            self._headloss_curves.add(key)
        elif curve_type == "VOLUME":
            self._volume_curves.add(key)
        elif curve_type == "EFFICIENCY":
            self._efficiency_curves.add(key)
        else:
            raise ValueError("curve_type must be HEAD, HEADLOSS, VOLUME, or EFFICIENCY")

    def add_curve(self, name, curve_type, xy_tuples_list):
        """
        Adds a curve to the water network model.

        Parameters
        ----------
        name : str
            Name of the curve.
        curve_type : str
            Type of curve. Options are HEAD, EFFICIENCY, VOLUME, HEADLOSS.
        xy_tuples_list : list of (x, y) tuples
            List of X-Y coordinate tuples on the curve.
        """
        assert (
            isinstance(name, str) and len(name) < 32 and name.find(" ") == -1
        ), "name must be a string with less than 32 characters and contain no spaces"
        assert isinstance(curve_type, (type(None), str)), "curve_type must be a string"
        assert isinstance(xy_tuples_list, (list, np.ndarray)), "xy_tuples_list must be a list of (x,y) tuples"

        curve = Curve(name, curve_type, xy_tuples_list)
        self[name] = curve

    def untyped_curves(self):
        """Generator to get all curves without type
        
        Yields
        ------
        name : str
            The name of the curve
        curve : Curve
            The untyped curve object    
            
        """
        defined = set(self._data.keys())
        untyped = defined.difference(
            self._pump_curves, self._efficiency_curves, self._headloss_curves, self._volume_curves
        )
        for key in untyped:
            yield key, self._data[key]

    @property
    def untyped_curve_names(self):
        """List of names of all curves without types"""
        defined = set(self._data.keys())
        untyped = defined.difference(
            self._pump_curves, self._efficiency_curves, self._headloss_curves, self._volume_curves
        )
        return list(untyped)

    def pump_curves(self):
        """Generator to get all pump curves
        
        Yields
        ------
        name : str
            The name of the curve
        curve : Curve
            The pump curve object    
            
        """
        for key in self._pump_curves:
            yield key, self._data[key]

    @property
    def pump_curve_names(self):
        """List of names of all pump curves"""
        return list(self._pump_curves)

    def efficiency_curves(self):
        """Generator to get all efficiency curves
        
        Yields
        ------
        name : str
            The name of the curve
        curve : Curve
            The efficiency curve object    
            
        """
        for key in self._efficiency_curves:
            yield key, self._data[key]

    @property
    def efficiency_curve_names(self):
        """List of names of all efficiency curves"""
        return list(self._efficiency_curves)

    def headloss_curves(self):
        """Generator to get all headloss curves
        
        Yields
        ------
        name : str
            The name of the curve
        curve : Curve
            The headloss curve object    
            
        """
        for key in self._headloss_curves:
            yield key, self._data[key]

    @property
    def headloss_curve_names(self):
        """List of names of all headloss curves"""
        return list(self._headloss_curves)

    def volume_curves(self):
        """Generator to get all volume curves
        
        Yields
        ------
        name : str
            The name of the curve
        curve : Curve
            The volume curve object    
            
        """
        for key in self._volume_curves:
            yield key, self._data[key]

    @property
    def volume_curve_names(self):
        """List of names of all volume curves"""
        return list(self._volume_curves)


class SourceRegistry(Registry):
    """A registry for sources."""

    def _finalize_(self, model):
        super()._finalize_(model)
        self._sources = None

    def __delitem__(self, key):
        try:
            if self._usage and key in self._usage and len(self._usage[key]) > 0:
                raise RuntimeError(
                    "cannot remove %s %s, still used by %s" % (self.__class__.__name__, key, self._usage[key])
                )
            elif key in self._usage:
                self._usage.pop(key)
            source = self._data.pop(key)
            self._pattern_reg.remove_usage(source.strength_timeseries.pattern_name, (source.name, "Source"))
            self._node_reg.remove_usage(source.node_name, (source.name, "Source"))
            return source
        except KeyError:
            # Do not raise an exception if there is no key of that name
            return


class NodeRegistry(Registry):
    """A registry for nodes."""

    def __init__(self, model):
        super(NodeRegistry, self).__init__(model)
        self._junctions = OrderedSet()
        self._reservoirs = OrderedSet()
        self._tanks = OrderedSet()

    def _finalize_(self, model):
        super()._finalize_(model)
        self._node_reg = None

    def __setitem__(self, key, value):
        if not isinstance(key, six.string_types):
            raise ValueError("Registry keys must be strings")
        self._data[key] = value
        if isinstance(value, Junction):
            self._junctions.add(key)
        elif isinstance(value, Tank):
            self._tanks.add(key)
        elif isinstance(value, Reservoir):
            self._reservoirs.add(key)

    def __delitem__(self, key):
        try:
            if self._usage and key in self._usage and len(self._usage[key]) > 0:
                raise RuntimeError(
                    "cannot remove %s %s, still used by %s" % (self.__class__.__name__, key, str(self._usage[key]))
                )
            elif key in self._usage:
                self._usage.pop(key)
            node = self._data.pop(key)
            self._junctions.discard(key)
            self._reservoirs.discard(key)
            self._tanks.discard(key)
            if isinstance(node, Junction):
                for pat_name in node.demand_timeseries_list.pattern_list():
                    if pat_name:
                        self._curve_reg.remove_usage(pat_name, (node.name, "Junction"))
            if isinstance(node, Reservoir) and node.head_pattern_name:
                self._curve_reg.remove_usage(node.head_pattern_name, (node.name, "Reservoir"))
            if isinstance(node, Tank) and node.vol_curve_name:
                self._curve_reg.remove_usage(node.vol_curve_name, (node.name, "Tank"))
            return node
        except KeyError:
            return

    def __call__(self, node_type=None):
        """
        Returns a generator to iterate over all nodes of a specific node type.
        If no node type is specified, the generator iterates over all nodes.

        Parameters
        ----------
        node_type: Node type
            Node type, options include
            wntr.network.model.Node,
            wntr.network.model.Junction,
            wntr.network.model.Reservoir,
            wntr.network.model.Tank, or None. Default = None.
            Note None and wntr.network.model.Node produce the same results.

        Returns
        -------
        A generator in the format (name, object).
        """
        if node_type == None:
            for node_name, node in self._data.items():
                yield node_name, node
        elif node_type == Junction:
            for node_name in self._junctions:
                yield node_name, self._data[node_name]
        elif node_type == Tank:
            for node_name in self._tanks:
                yield node_name, self._data[node_name]
        elif node_type == Reservoir:
            for node_name in self._reservoirs:
                yield node_name, self._data[node_name]
        else:
            raise RuntimeError("node_type, " + str(node_type) + ", not recognized.")

    def add_junction(
        self,
        name,
        base_demand=0.0,
        demand_pattern=None,
        elevation=0.0,
        coordinates=None,
        demand_category=None,
        emitter_coeff=None,
        initial_quality=None,
    ):
        """
        Adds a junction to the water network model.

        Parameters
        -------------------
        name : str
            Name of the junction.
        base_demand : float
            Base demand at the junction.
        demand_pattern : str or Pattern
            Name of the demand pattern or the Pattern object
        elevation : float
            Elevation of the junction.
        coordinates : tuple of floats, optional
            X-Y coordinates of the node location.
        demand_category : str, optional
            Category to the **base** demand
        emitter_coeff : float, optional
            Emitter coefficient
        initial_quality : float, optional
            Initial quality at this junction
        """
        assert (
            isinstance(name, str) and len(name) < 32 and name.find(" ") == -1
        ), "name must be a string with less than 32 characters and contain no spaces"
        assert isinstance(base_demand, (int, float)), "base_demand must be a float"
        assert isinstance(
            demand_pattern, (type(None), str, PatternRegistry.DefaultPattern, Pattern)
        ), "demand_pattern must be a string or Pattern"
        assert isinstance(elevation, (int, float)), "elevation must be a float"
        assert isinstance(coordinates, (type(None), (tuple, list,))), "coordinates must be a tuple"
        assert isinstance(demand_category, (type(None), str)), "demand_category must be a string"
        assert isinstance(emitter_coeff, (type(None), int, float)), "emitter_coeff must be a float"
        assert isinstance(initial_quality, (type(None), int, float)), "initial_quality must be a float"

        base_demand = float(base_demand)
        elevation = float(elevation)

        junction = Junction(name, self)
        junction.elevation = elevation
        junction.add_demand(base_demand, demand_pattern, demand_category)
        self[name] = junction
        if coordinates is not None:
            junction.coordinates = coordinates
        if emitter_coeff is not None:
            junction.emitter_coefficient = emitter_coeff
        if initial_quality is not None:
            junction.initial_quality = initial_quality

    def add_tank(
        self,
        name,
        elevation=0.0,
        init_level=3.048,
        min_level=0.0,
        max_level=6.096,
        diameter=15.24,
        min_vol=0.0,
        vol_curve=None,
        overflow=False,
        coordinates=None,
    ):
        """
        Adds a tank to the water network model.

        Parameters
        -------------------
        name : str
            Name of the tank.
        elevation : float
            Elevation at the tank.
        init_level : float
            Initial tank level.
        min_level : float
            Minimum tank level.
        max_level : float
            Maximum tank level.
        diameter : float
            Tank diameter of a cylindrical tank (only used when the volume 
            curve is None)
        min_vol : float
            Minimum tank volume (only used when the volume curve is None)
        vol_curve : str, optional
            Name of a volume curve. The volume curve overrides the tank diameter
            and minimum volume.
        overflow : bool, optional
            Overflow indicator (Always False for the WNTRSimulator)
        coordinates : tuple of floats, optional
            X-Y coordinates of the node location.
            
        """
        assert (
            isinstance(name, str) and len(name) < 32 and name.find(" ") == -1
        ), "name must be a string with less than 32 characters and contain no spaces"
        assert isinstance(elevation, (int, float)), "elevation must be a float"
        assert isinstance(init_level, (int, float)), "init_level must be a float"
        assert isinstance(min_level, (int, float)), "min_level must be a float"
        assert isinstance(max_level, (int, float)), "max_level must be a float"
        assert isinstance(diameter, (int, float)), "diameter must be a float"
        assert isinstance(min_vol, (int, float)), "min_vol must be a float"
        assert isinstance(vol_curve, (type(None), str)), "vol_curve must be a string"
        assert isinstance(overflow, (type(None), str, bool, int)), "overflow must be a bool, 'YES' or 'NO, or 0 or 1"
        assert isinstance(coordinates, (type(None), (tuple,list,))), "coordinates must be a tuple"
        
        elevation = float(elevation)
        init_level = float(init_level)
        min_level = float(min_level)
        max_level = float(max_level)
        diameter = float(diameter)
        min_vol = float(min_vol)
        if init_level < min_level:
            raise ValueError("Initial tank level must be greater than or equal to the tank minimum level.")
        if init_level > max_level:
            raise ValueError("Initial tank level must be less than or equal to the tank maximum level.")
        if vol_curve is not None and vol_curve != "*":
            if not isinstance(vol_curve, six.string_types):
                raise ValueError("Volume curve name must be a string")
            elif not vol_curve in self._curve_reg.volume_curve_names:
                raise ValueError(
                    "The volume curve "
                    + vol_curve
                    + " is not one of the curves in the "
                    + "list of volume curves. Valid volume curves are:"
                    + str(self._curve_reg.volume_curve_names)
                )
            vcurve = np.array(self._curve_reg[vol_curve].points)
            if min_level < vcurve[0, 0]:
                raise ValueError(
                    (
                        "The volume curve "
                        + vol_curve
                        + " has a minimum value ({0:5.2f}) \n"
                        + 'greater than the minimum level for tank "'
                        + name
                        + '" ({1:5.2f})\n'
                        + "please correct the user input."
                    ).format(vcurve[0, 0], min_level)
                )
            elif max_level > vcurve[-1, 0]:
                raise ValueError(
                    (
                        "The volume curve "
                        + vol_curve
                        + " has a maximum value ({0:5.2f}) \n"
                        + 'less than the maximum level for tank "'
                        + name
                        + '" ({1:5.2f})\n'
                        + "please correct the user input."
                    ).format(vcurve[-1, 0], max_level)
                )

        tank = Tank(name, self)
        tank.elevation = elevation
        tank.init_level = init_level
        tank.min_level = min_level
        tank.max_level = max_level
        tank.diameter = diameter
        tank.min_vol = min_vol
        tank.vol_curve_name = vol_curve
        tank.overflow = overflow
        self[name] = tank
        if coordinates is not None:
            tank.coordinates = coordinates

    def add_reservoir(self, name, base_head=0.0, head_pattern=None, coordinates=None):
        """
        Adds a reservoir to the water network model.

        Parameters
        ----------
        name : str
            Name of the reservoir.
        base_head : float, optional
            Base head at the reservoir.
        head_pattern : str, optional
            Name of the head pattern.
        coordinates : tuple of floats, optional
            X-Y coordinates of the node location.
        
        """
        assert (
            isinstance(name, str) and len(name) < 32 and name.find(" ") == -1
        ), "name must be a string with less than 32 characters and contain no spaces"
        assert isinstance(base_head, (int, float)), "base_head must be float"
        assert isinstance(head_pattern, (type(None), str)), "head_pattern must be a string"
        assert isinstance(coordinates, (type(None), (tuple, list))), "coordinates must be a tuple"

        base_head = float(base_head)

        reservoir = Reservoir(name, self)
        reservoir.base_head = base_head
        reservoir.head_pattern_name = head_pattern
        self[name] = reservoir
        if coordinates is not None:
            reservoir.coordinates = coordinates

    @property
    def junction_names(self):
        """List of names of all junctions"""
        return self._junctions

    @property
    def tank_names(self):
        """List of names of all junctions"""
        return self._tanks

    @property
    def reservoir_names(self):
        """List of names of all junctions"""
        return self._reservoirs

    def junctions(self):
        """Generator to get all junctions
        
        Yields
        ------
        name : str
            The name of the junction
        node : Junction
            The junction object    
            
        """
        for node_name in self._junctions:
            yield node_name, self._data[node_name]

    def tanks(self):
        """Generator to get all tanks
        
        Yields
        ------
        name : str
            The name of the tank
        node : Tank
            The tank object    
            
        """
        for node_name in self._tanks:
            yield node_name, self._data[node_name]

    def reservoirs(self):
        """Generator to get all reservoirs
        
        Yields
        ------
        name : str
            The name of the reservoir
        node : Reservoir
            The reservoir object    
            
        """
        for node_name in self._reservoirs:
            yield node_name, self._data[node_name]


class LinkRegistry(Registry):
    """A registry for links."""

    __subsets = [
        "_pipes",
        "_pumps",
        "_head_pumps",
        "_power_pumps",
        "_prvs",
        "_psvs",
        "_pbvs",
        "_tcvs",
        "_fcvs",
        "_gpvs",
        "_valves",
    ]

    def __init__(self, model):
        super(LinkRegistry, self).__init__(model)
        self._pipes = OrderedSet()
        self._pumps = OrderedSet()
        self._head_pumps = OrderedSet()
        self._power_pumps = OrderedSet()
        self._prvs = OrderedSet()
        self._psvs = OrderedSet()
        self._pbvs = OrderedSet()
        self._tcvs = OrderedSet()
        self._fcvs = OrderedSet()
        self._gpvs = OrderedSet()
        self._valves = OrderedSet()

    def _finalize_(self, model):
        super()._finalize_(model)
        self._link_reg = None

    def __setitem__(self, key, value):
        if not isinstance(key, six.string_types):
            raise ValueError("Registry keys must be strings")
        self._data[key] = value
        if isinstance(value, Pipe):
            self._pipes.add(key)
        elif isinstance(value, Pump):
            self._pumps.add(key)
            if isinstance(value, HeadPump):
                self._head_pumps.add(key)
            elif isinstance(value, PowerPump):
                self._power_pumps.add(key)
        elif isinstance(value, Valve):
            self._valves.add(key)
            if isinstance(value, PRValve):
                self._prvs.add(key)
            elif isinstance(value, PSValve):
                self._psvs.add(key)
            elif isinstance(value, PBValve):
                self._pbvs.add(key)
            elif isinstance(value, TCValve):
                self._tcvs.add(key)
            elif isinstance(value, FCValve):
                self._fcvs.add(key)
            elif isinstance(value, GPValve):
                self._gpvs.add(key)

    def __delitem__(self, key):
        try:
            if self._usage and key in self._usage and len(self._usage[key]) > 0:
                raise RuntimeError(
                    "cannot remove %s %s, still used by %s", self.__class__.__name__, key, self._usage[key]
                )
            elif key in self._usage:
                self._usage.pop(key)
            link = self._data.pop(key)
            self._node_reg.remove_usage(link.start_node_name, (link.name, link.link_type))
            self._node_reg.remove_usage(link.end_node_name, (link.name, link.link_type))
            if isinstance(link, GPValve):
                self._curve_reg.remove_usage(link.headloss_curve_name, (link.name, "Valve"))
            if isinstance(link, Pump):
                self._curve_reg.remove_usage(link.speed_pattern_name, (link.name, "Pump"))
            if isinstance(link, HeadPump):
                self._curve_reg.remove_usage(link.pump_curve_name, (link.name, "Pump"))
            for ss in self.__subsets:
                # Go through the _pipes, _prvs, ..., and remove this link
                getattr(self, ss).discard(key)
            return link
        except KeyError:
            return

    def __call__(self, link_type=None):
        """
        Returns a generator to iterate over all nodes of a specific node type.
        If no node type is specified, the generator iterates over all nodes.

        Parameters
        ----------
        node_type: Node type
            Node type, options include
            wntr.network.model.Node,
            wntr.network.model.Junction,
            wntr.network.model.Reservoir,
            wntr.network.model.Tank, or None. Default = None.
            Note None and wntr.network.model.Node produce the same results.

        Returns
        -------
        A generator in the format (name, object).
        """
        if link_type == None:
            for name, node in self._data.items():
                yield name, node
        elif link_type == Pipe:
            for name in self._pipes:
                yield name, self._data[name]
        elif link_type == Pump:
            for name in self._pumps:
                yield name, self._data[name]
        elif link_type == Valve:
            for name in self._valves:
                yield name, self._data[name]
        else:
            raise RuntimeError("link_type, " + str(link_type) + ", not recognized.")

    def add_pipe(
        self,
        name,
        start_node_name,
        end_node_name,
        length=304.8,
        diameter=0.3048,
        roughness=100,
        minor_loss=0.0,
        initial_status="OPEN",
        check_valve=False,
    ):
        """
        Adds a pipe to the water network model.

        Parameters
        ----------
        name : str
            Name of the pipe.
        start_node_name : str
             Name of the start node.
        end_node_name : str
             Name of the end node.
        length : float, optional
            Length of the pipe.
        diameter : float, optional
            Diameter of the pipe.
        roughness : float, optional
            Pipe roughness coefficient.
        minor_loss : float, optional
            Pipe minor loss coefficient.
        initial_status : str, optional
            Pipe initial status. Options are 'OPEN' or 'CLOSED'.
        check_valve : bool, optional
            True if the pipe has a check valve.
            False if the pipe does not have a check valve.
        
        """
        assert (
            isinstance(name, str) and len(name) < 32 and name.find(" ") == -1
        ), "name must be a string with less than 32 characters and contain no spaces"
        assert (
            isinstance(start_node_name, str) and len(start_node_name) < 32 and start_node_name.find(" ") == -1
        ), "start_node_name must be a string with less than 32 characters and contain no spaces"
        assert (
            isinstance(end_node_name, str) and len(end_node_name) < 32 and end_node_name.find(" ") == -1
        ), "end_node_name must be a string with less than 32 characters and contain no spaces"
        assert isinstance(length, (int, float)), "length must be a float"
        assert isinstance(diameter, (int, float)), "diameter must be a float"
        assert isinstance(roughness, (int, float)), "roughness must be a float"
        assert isinstance(minor_loss, (int, float)), "minor_loss must be a float"
        assert isinstance(initial_status, (int, str, LinkStatus)), "initial_status must be an int, string or LinkStatus"
        assert isinstance(check_valve, (int, str, LinkStatus, type(None))), "initial_status must be an int, string, LinkStatus, or None type"

        
        length = float(length)
        diameter = float(diameter)
        roughness = float(roughness)
        minor_loss = float(minor_loss)
        if isinstance(initial_status, str):
            initial_status = LinkStatus[initial_status]

        pipe = Pipe(name, start_node_name, end_node_name, self)
        pipe.length = length
        pipe.diameter = diameter
        pipe.roughness = roughness
        pipe.minor_loss = minor_loss
        pipe.initial_status = initial_status
        pipe._user_status = initial_status
        pipe.check_valve = check_valve
        self[name] = pipe

    def add_pump(
        self,
        name,
        start_node_name,
        end_node_name,
        pump_type="POWER",
        pump_parameter=50.0,
        speed=1.0,
        pattern=None,
        initial_status="OPEN",
    ):
        """
        Adds a pump to the water network model.

        Parameters
        ----------
        name : string
            Name of the pump.
        start_node_name : string
             Name of the start node.
        end_node_name : string
             Name of the end node.
        pump_type : string, optional
            Type of information provided for a pump. Options are 'POWER' or 'HEAD'.
        pump_parameter : float or string
            For a POWER pump, the pump power (float).
            For a HEAD pump, the head curve name (string).
        speed: float
            Relative speed setting (1.0 is normal speed)
        pattern: string
            Name of the speed pattern
        initial_status: str or LinkStatus
            Pump initial status. Options are 'OPEN' or 'CLOSED'.
        
        """
        assert (
            isinstance(name, str) and len(name) < 32 and name.find(" ") == -1
        ), "name must be a string with less than 32 characters and contain no spaces"
        assert (
            isinstance(start_node_name, str) and len(start_node_name) < 32 and start_node_name.find(" ") == -1
        ), "start_node_name must be a string with less than 32 characters and contain no spaces"
        assert (
            isinstance(end_node_name, str) and len(end_node_name) < 32 and end_node_name.find(" ") == -1
        ), "end_node_name must be a string with less than 32 characters and contain no spaces"
        assert isinstance(pump_type, str), "pump_type must be a string"
        assert isinstance(pump_parameter, (int, float, str)), "pump_parameter must be a float or string"
        assert isinstance(speed, (int, float)), "speed must be a float"
        assert isinstance(pattern, (type(None), str)), "pattern must be a string"
        assert isinstance(initial_status, (int, str, LinkStatus)), "initial_status must be an int, string or LinkStatus"

        if isinstance(initial_status, str):
            initial_status = LinkStatus[initial_status]
        if pump_type.upper() == "POWER":
            pump = PowerPump(name, start_node_name, end_node_name, self)
            pump.power = pump_parameter
        elif pump_type.upper() == "HEAD":
            pump = HeadPump(name, start_node_name, end_node_name, self)
            pump.pump_curve_name = pump_parameter
        else:
            raise ValueError('pump_type must be "POWER" or "HEAD"')
        pump.base_speed = speed
        pump.initial_status = initial_status
        pump.speed_pattern_name = pattern
        self[name] = pump

    def add_valve(
        self,
        name,
        start_node_name,
        end_node_name,
        diameter=0.3048,
        valve_type="PRV",
        minor_loss=0.0,
        initial_setting=0.0,
        initial_status="ACTIVE",
    ):
        """
        Adds a valve to the water network model.

        Parameters
        ----------
        name : string
            Name of the valve.
        start_node_name : string
             Name of the start node.
        end_node_name : string
             Name of the end node.
        diameter : float, optional
            Diameter of the valve.
        valve_type : string, optional
            Type of valve. Options are 'PRV', 'PSV', 'PBV', 'FCV', 'TCV', and 'GPV'
        minor_loss : float, optional
            Pipe minor loss coefficient.
        initial_setting : float or string, optional
            Valve initial setting.
            Pressure setting for PRV, PSV, or PBV. 
            Flow setting for FCV. 
            Loss coefficient for TCV.
            Name of headloss curve for GPV.
        initial_status: string or LinkStatus
            Valve initial status. Options are 'OPEN',  'CLOSED', or 'ACTIVE'
            
        """
        assert (
            isinstance(name, str) and len(name) < 32 and name.find(" ") == -1
        ), "name must be a string with less than 32 characters and contain no spaces"
        assert (
            isinstance(start_node_name, str) and len(start_node_name) < 32 and start_node_name.find(" ") == -1
        ), "start_node_name must be a string with less than 32 characters and contain no spaces"
        assert (
            isinstance(end_node_name, str) and len(end_node_name) < 32 and end_node_name.find(" ") == -1
        ), "end_node_name must be a string with less than 32 characters and contain no spaces"
        assert isinstance(diameter, (int, float)), "diameter must be a float"
        assert isinstance(valve_type, str), "valve_type must be a string"
        assert isinstance(minor_loss, (int, float)), "minor_loss must be a float"
        assert isinstance(initial_setting, (int, float, str)), "initial_setting must be a float or string"
        assert isinstance(initial_status, (str, LinkStatus)), "initial_status must be a string or LinkStatus"

        if isinstance(initial_status, str):
            initial_status = LinkStatus[initial_status]
        start_node = self._node_reg[start_node_name]
        end_node = self._node_reg[end_node_name]

        valve_type = valve_type.upper()

        # A PRV, PSV or FCV cannot be directly connected to a reservoir or tank (use a length of pipe to separate the two)
        if valve_type in ["PRV", "PSV", "FCV"]:
            if type(start_node) == Tank or type(end_node) == Tank:
                msg = (
                    "%ss cannot be directly connected to a tank.  Add a pipe to separate the valve from the tank."
                    % valve_type
                )
                logger.error(msg)
                raise RuntimeError(msg)
            if type(start_node) == Reservoir or type(end_node) == Reservoir:
                msg = (
                    "%ss cannot be directly connected to a reservoir.  Add a pipe to separate the valve from the reservoir."
                    % valve_type
                )
                logger.error(msg)
                raise RuntimeError(msg)

        # TODO check the following: PRVs cannot share the same downstream node or be linked in series

        # TODO check the following: Two PSVs cannot share the same upstream node or be linked in series

        # TODO check the following: A PSV cannot be connected to the downstream node of a PRV

        if valve_type == "PRV":
            valve = PRValve(name, start_node_name, end_node_name, self)
            valve.initial_setting = initial_setting
            valve._setting = initial_setting
        elif valve_type == "PSV":
            valve = PSValve(name, start_node_name, end_node_name, self)
            valve.initial_setting = initial_setting
            valve._setting = initial_setting
        elif valve_type == "PBV":
            valve = PBValve(name, start_node_name, end_node_name, self)
            valve.initial_setting = initial_setting
            valve._setting = initial_setting
        elif valve_type == "FCV":
            valve = FCValve(name, start_node_name, end_node_name, self)
            valve.initial_setting = initial_setting
            valve._setting = initial_setting
        elif valve_type == "TCV":
            valve = TCValve(name, start_node_name, end_node_name, self)
            valve.initial_setting = initial_setting
            valve._setting = initial_setting
        elif valve_type == "GPV":
            valve = GPValve(name, start_node_name, end_node_name, self)
            valve.headloss_curve_name = initial_setting
        valve.initial_status = initial_status
        valve.diameter = diameter
        valve.minor_loss = minor_loss
        self[name] = valve

    def check_valves(self):
        """Generator to get all pipes with check valves
        
        Yields
        ------
        name : str
            The name of the pipe
        link : Pipe
            The pipe object    
            
        """
        for name in self._pipes:
            if self._data[name].check_valve:
                yield name

    @property
    def pipe_names(self):
        """A list of all pipe names"""
        return self._pipes

    @property
    def valve_names(self):
        """A list of all valve names"""
        return self._valves

    @property
    def pump_names(self):
        """A list of all pump names"""
        return self._pumps

    @property
    def head_pump_names(self):
        """A list of all head pump names"""
        return self._head_pumps

    @property
    def power_pump_names(self):
        """A list of all power pump names"""
        return self._power_pumps

    @property
    def prv_names(self):
        """A list of all prv names"""
        return self._prvs

    @property
    def psv_names(self):
        """A list of all psv names"""
        return self._psvs

    @property
    def pbv_names(self):
        """A list of all pbv names"""
        return self._pbvs

    @property
    def tcv_names(self):
        """A list of all tcv names"""
        return self._tcvs

    @property
    def fcv_names(self):
        """A list of all fcv names"""
        return self._fcvs

    @property
    def gpv_names(self):
        """A list of all gpv names"""
        return self._gpvs

    def pipes(self):
        """Generator to get all pipes
        
        Yields
        ------
        name : str
            The name of the pipe
        link : Pipe
            The pipe object    
            
        """
        for name in self._pipes:
            yield name, self._data[name]

    def pumps(self):
        """Generator to get all pumps
        
        Yields
        ------
        name : str
            The name of the pump
        link : Pump
            The pump object    
            
        """
        for name in self._pumps:
            yield name, self._data[name]

    def valves(self):
        """Generator to get all valves
        
        Yields
        ------
        name : str
            The name of the valve
        link : Valve
            The valve object    
            
        """
        for name in self._valves:
            yield name, self._data[name]

    def head_pumps(self):
        """Generator to get all head pumps
        
        Yields
        ------
        name : str
            The name of the pump
        link : HeadPump
            The pump object    
            
        """
        for name in self._head_pumps:
            yield name, self._data[name]

    def power_pumps(self):
        """Generator to get all power pumps
        
        Yields
        ------
        name : str
            The name of the pump
        link : PowerPump
            The pump object    
            
        """
        for name in self._power_pumps:
            yield name, self._data[name]

    def prvs(self):
        """Generator to get all PRVs
        
        Yields
        ------
        name : str
            The name of the valve
        link : PRValve
            The valve object
            
        """
        for name in self._prvs:
            yield name, self._data[name]

    def psvs(self):
        """Generator to get all PSVs
        
        Yields
        ------
        name : str
            The name of the valve
        link : PSValve
            The valve object
            
        """
        for name in self._psvs:
            yield name, self._data[name]

    def pbvs(self):
        """Generator to get all PBVs
        
        Yields
        ------
        name : str
            The name of the valve
        link : PBValve
            The valve object
            
        """
        for name in self._pbvs:
            yield name, self._data[name]

    def tcvs(self):
        """Generator to get all TCVs
        
        Yields
        ------
        name : str
            The name of the valve
        link : TCValve
            The valve object
            
        """
        for name in self._tcvs:
            yield name, self._data[name]

    def fcvs(self):
        """Generator to get all FCVs
        
        Yields
        ------
        name : str
            The name of the valve
        link : FCValve
            The valve object
            
        """
        for name in self._fcvs:
            yield name, self._data[name]

    def gpvs(self):
        """Generator to get all GPVs
        
        Yields
        ------
        name : str
            The name of the valve
        link : GPValve
            The valve object
            
        """
        for name in self._gpvs:
            yield name, self._data[name]




"""
The wntr.library package contains classes to help define water network models
"""
from .model_library import ModelLibrary
from .demand_library import DemandPatternLibrary
from . import msx

model_library = ModelLibrary()
demand_library = DemandPatternLibrary()
reaction_library = msx.MsxLibrary()



import unittest
from os.path import abspath, dirname, join
import sys, platform

import wntr

if 'darwin' in sys.platform.lower() and 'arm' in platform.platform().lower():
    skip_v2_tests_on_arm = True
else:
    skip_v2_tests_on_arm = False

testdir = dirname(abspath(str(__file__)))
datadir = join(testdir, "networks_for_testing")
netdir = join(testdir, "..", "..", "examples", "networks")


class TestHealthImpactsMetric(unittest.TestCase):
    """
    Compare the following results to WST impact files using TSG file
    121          SETPOINT      100000          0                86400
    """

    def test_mass_consumed(self):
        inp_file = join(netdir, "Net3.inp")
        if skip_v2_tests_on_arm: self.skipTest('skipped test due to skip_tests_flag')

        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.hydraulic_timestep = 15*60
        wn.options.time.quality_timestep = 15*60
        wn.options.time.report_timestep = 15*60

        wn.options.quality.parameter = "CHEMICAL"
        newpat = wntr.network.elements.Pattern.binary_pattern(
            "NewPattern",
            0,
            24 * 3600,
            wn.options.time.pattern_timestep,
            wn.options.time.duration,
        )
        wn.add_pattern(newpat.name, newpat)
        wn.add_source("Source1", "121", "SETPOINT", 100, "NewPattern")

        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim(version=2.0)

        demand = results.node["demand"].loc[:, wn.junction_name_list]
        quality = results.node["quality"].loc[:, wn.junction_name_list]

        MC = wntr.metrics.mass_contaminant_consumed(demand, quality)
        MC_timeseries = MC.sum(axis=1)
        MC_cumsum = MC_timeseries.cumsum()
        # MC_timeseries.to_csv('MC.txt')

        expected = float(39069900000 / 1000000)  # hour 2
        error = abs((MC_cumsum[2 * 3600] - expected) / expected)
        self.assertLess(error, 0.01)  # 1% error

        expected = float(1509440000000 / 1000000)  # hour 12
        error = abs((MC_cumsum[12 * 3600] - expected) / expected)
        self.assertLess(error, 0.01)  # 1% error

    def test_volume_consumed(self):
        if skip_v2_tests_on_arm: self.skipTest('skipped test due to skip_tests_flag')

        inp_file = join(netdir, "Net3.inp")

        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.hydraulic_timestep = 15*60
        wn.options.time.quality_timestep = 15*60
        wn.options.time.report_timestep = 15*60
        
        wn.options.quality.parameter = "CHEMICAL"
        newpat = wntr.network.elements.Pattern.binary_pattern(
            "NewPattern",
            0,
            24 * 3600,
            wn.options.time.pattern_timestep,
            wn.options.time.duration,
        )
        wn.add_pattern(newpat.name, newpat)
        wn.add_source("Source1", "121", "SETPOINT", 100, "NewPattern")

        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim(version=2.0)

        demand = results.node["demand"].loc[:, wn.junction_name_list]
        quality = results.node["quality"].loc[:, wn.junction_name_list]

        VC = wntr.metrics.volume_contaminant_consumed(demand, quality, 0)
        VC_timeseries = VC.sum(axis=1)
        VC_cumsum = VC_timeseries.cumsum()
        # VC_timeseries.to_csv('VC.txt')

        expected = float(156760 / 264.172)  # hour 2, convert gallons to m3
        error = abs((VC_cumsum[2 * 3600] - expected) / expected)
        self.assertLess(error, 0.02)  # 2% error

        expected = float(4867920 / 264.172)  # hour 12, convert gallons to m3
        error = abs((VC_cumsum[12 * 3600] - expected) / expected)
        self.assertLess(error, 0.01)  # 1% error

    def test_extent_contaminated(self):
        if skip_v2_tests_on_arm: self.skipTest('skipped test due to skip_tests_flag')

        inp_file = join(netdir, "Net3.inp")

        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.hydraulic_timestep = 15*60
        wn.options.time.quality_timestep = 15*60
        wn.options.time.report_timestep = 15*60
        
        wn.options.quality.parameter = "CHEMICAL"
        newpat = wntr.network.elements.Pattern.binary_pattern(
            "NewPattern",
            0,
            24 * 3600,
            wn.options.time.pattern_timestep,
            wn.options.time.duration,
        )
        wn.add_pattern(newpat.name, newpat)
        wn.add_source("Source1", "121", "SETPOINT", 100, "NewPattern")

        sim = wntr.sim.EpanetSimulator(wn)
        results = sim.run_sim(version=2.0)

        quality = results.node["quality"]
        flowrate = results.link["flowrate"].loc[:, wn.pipe_name_list]

        EC = wntr.metrics.extent_contaminant(quality, flowrate, wn, 0)

        expected = float(80749.9 * 0.3048)  # hour 2
        error = abs((EC[2 * 3600] - expected) / expected)
        self.assertLess(error, 0.01)  # 1% error

        expected = float(135554 * 0.3048)  # hour 12
        error = abs((EC[12 * 3600] - expected) / expected)
        self.assertLess(error, 0.01)  # 1% error


if __name__ == "__main__":
    unittest.main()



import unittest
import warnings
from os.path import abspath, dirname, join

import wntr

testdir = dirname(abspath(str(__file__)))
test_datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")

class TestPatternStart(unittest.TestCase):

    def test_pattern_start(self):

        inp_file = join(ex_datadir, "Net1.inp")

        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.pattern_start = 0

        sim = wntr.sim.EpanetSimulator(wn)
        epa_demand_0 = sim.run_sim().node['demand'].loc[:,wn.junction_name_list]

        sim = wntr.sim.WNTRSimulator(wn)
        wntr_demand_0 = sim.run_sim().node['demand'].loc[:,wn.junction_name_list]

        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.time.pattern_start = 3*3600

        sim = wntr.sim.EpanetSimulator(wn)
        epa_demand_3 = sim.run_sim().node['demand'].loc[:,wn.junction_name_list]

        sim = wntr.sim.WNTRSimulator(wn)
        wntr_demand_3 = sim.run_sim().node['demand'].loc[:,wn.junction_name_list]

        diff_demand_0 = abs(epa_demand_0 - wntr_demand_0).sum().sum()
        self.assertLess(diff_demand_0, 1e-5)

        diff_demand_3 = abs(epa_demand_3 - wntr_demand_3).sum().sum()
        self.assertLess(diff_demand_3, 1e-5)

        epa_demand_3.index = epa_demand_3.index + 3*3600
        diff_epa_shifted = abs(epa_demand_0 - epa_demand_3).sum().sum()
        self.assertLess(diff_epa_shifted, 1e-5)

        wntr_demand_3.index = wntr_demand_3.index + 3*3600
        diff_wntr_shifted = abs(wntr_demand_0 - wntr_demand_3).sum().sum()
        self.assertLess(diff_wntr_shifted, 1e-5)

class TestDemandMultiplier(unittest.TestCase):
    
    def test_demand_multiplier(self):

        node_name = "147"
        time = 3600

        inp_file = join(ex_datadir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        node = wn.get_node(node_name)

        # Demand multiplier = 1
        wn.options.hydraulic.demand_multiplier = 1
        sim = wntr.sim.EpanetSimulator(wn)
        epanet_results1 = sim.run_sim()
        sim = wntr.sim.WNTRSimulator(wn)
        wntr_results1 = sim.run_sim()

        expected_demand1 = node.demand_timeseries_list.at(time)
        epanet_actual_demand1 = epanet_results1.node["demand"].loc[time, node_name]
        wntr_actual_demand1 = wntr_results1.node["demand"].loc[time, node_name]

        self.assertGreater(
            expected_demand1, 0
        )  # make sure expected demand is greater than 0

        self.assertAlmostEqual(expected_demand1, epanet_actual_demand1, 8)
        self.assertAlmostEqual(expected_demand1, wntr_actual_demand1, 8)
        self.assertAlmostEqual(epanet_actual_demand1, wntr_actual_demand1, 8)

        # Increase demand multiplier to 1.5
        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.hydraulic.demand_multiplier = 1.5
        sim = wntr.sim.EpanetSimulator(wn)
        epanet_results2 = sim.run_sim()
        sim = wntr.sim.WNTRSimulator(wn)
        wntr_results2 = sim.run_sim()

        expected_demand2 = node.demand_timeseries_list.at(time, multiplier=1.5)
        epanet_actual_demand2 = epanet_results2.node["demand"].loc[time, node_name]
        wntr_actual_demand2 = wntr_results2.node["demand"].loc[time, node_name]

        self.assertAlmostEqual(expected_demand1 * 1.5, epanet_actual_demand2, 8)

        self.assertAlmostEqual(expected_demand2, epanet_actual_demand2, 8)
        self.assertAlmostEqual(expected_demand2, wntr_actual_demand2, 8)
        self.assertAlmostEqual(epanet_actual_demand2, wntr_actual_demand2, 8)

        # Decrease demand multiplier to 0.5
        wn = wntr.network.WaterNetworkModel(inp_file)
        wn.options.hydraulic.demand_multiplier = 0.5
        sim = wntr.sim.EpanetSimulator(wn)
        epanet_results2 = sim.run_sim()
        sim = wntr.sim.WNTRSimulator(wn)
        wntr_results2 = sim.run_sim()

        expected_demand2 = node.demand_timeseries_list.at(time, multiplier=0.5)
        epanet_actual_demand2 = epanet_results2.node["demand"].loc[time, node_name]
        wntr_actual_demand2 = wntr_results2.node["demand"].loc[time, node_name]

        self.assertAlmostEqual(expected_demand1 * 0.5, epanet_actual_demand2, 8)

        self.assertAlmostEqual(expected_demand2, epanet_actual_demand2, 8)
        self.assertAlmostEqual(expected_demand2, wntr_actual_demand2, 8)
        self.assertAlmostEqual(epanet_actual_demand2, wntr_actual_demand2, 8)


if __name__ == "__main__":
    unittest.main()



import unittest
import warnings
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
import wntr
import wntr.msx
import wntr.epanet.msx

testdir = dirname(abspath(str(__file__)))
test_network_dir = join(testdir, "networks_for_testing")
inp_filename = join(test_network_dir, "msx_example.inp")
msx_filename = join(test_network_dir, "msx_example.msx")


class TestEpanetMSXSim(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        pass

    @classmethod
    def tearDownClass(self):
        pass

    def test_msx_sim(self):
        wn = wntr.network.WaterNetworkModel(inp_file_name=inp_filename)
        wn.add_msx_model(msx_filename=msx_filename)
        sim = wntr.sim.EpanetSimulator(wn)
        msx_model = wntr.msx.MsxModel(msx_file_name=msx_filename)
        
        # run sim
        res = sim.run_sim()
        
        # check results object keys
        for species in wn.msx.species_name_list:
            assert species in res.node.keys()
            assert species in res.link.keys()
            
        
        # sanity check at test point
        expected = 10.032905  # Node 'C' at time 136800 for AStot
        error = abs(
            (res.node["AStot"].loc[136800, "C"] - expected) / expected
        )
        self.assertLess(error, 0.0001)  # 0.01% error


if __name__ == "__main__":
    unittest.main(verbosity=2)



# coding: utf-8
"""
The wntr.epanet.msx.toolkit module is a Python extension for the EPANET-MSX
Programmers Toolkit DLLs.

.. note::
    
    Code in this section is based on code from "EPANET-MSX-Python-wrapper",
    licensed under the BSD license. See LICENSE.md for details.
"""
import ctypes
import logging
import os
import os.path
import platform
import sys
from typing import Union

if sys.version_info[0:2] <= (3, 11):
    from pkg_resources import resource_filename
else:
    from importlib.resources import files

from wntr.epanet.msx.enums import TkObjectType, TkSourceType

from ..toolkit import ENepanet
from .exceptions import (MSX_ERROR_CODES, EpanetMsxException, MSXKeyError,
                         MSXValueError)

logger = logging.getLogger(__name__)

epanet_toolkit = "wntr.epanet.toolkit"

class MSXepanet(ENepanet):
    def __init__(self, inpfile="", rptfile="", binfile="", msxfile=""):

        self.ENlib = None
        self.errcode = 0
        self.errcodelist = []
        self.cur_time = 0

        self.Warnflag = False
        self.Errflag = False
        self.fileLoaded = False

        self.inpfile = inpfile
        self.rptfile = rptfile
        self.binfile = binfile
        self.msxfile = msxfile

        try:
            if sys.version_info[0:2] <= (3, 11):

                if os.name in ["nt", "dos"]:
                    libepanet = resource_filename(epanet_toolkit, "libepanet/windows-x64/epanet2.dll")
                    libmsx = resource_filename(epanet_toolkit, "libepanet/windows-x64/epanetmsx.dll")
                elif sys.platform in ["darwin"]:
                    if 'arm' in platform.platform().lower():
                        libepanet = resource_filename(epanet_toolkit, "libepanet/darwin-arm/libepanet2.dylib")
                        libmsx = resource_filename(epanet_toolkit, "libepanet/darwin-arm/libepanetmsx.dylib")
                    else:
                        libepanet = resource_filename(epanet_toolkit, "libepanet/darwin-x64/libepanet2.dylib")
                        libmsx = resource_filename(epanet_toolkit, "libepanet/darwin-x64/libepanetmsx.dylib")
                else:
                    libepanet = resource_filename(epanet_toolkit, "libepanet/linux-x64/libepanet2.so")
                    libmsx = resource_filename(epanet_toolkit, "libepanet/linux-x64/libepanetmsx.so")

                dylib_dir = os.environ.get('DYLD_FALLBACK_LIBRARY_PATH','')
                if dylib_dir != '':
                    if 'arm' in platform.platform().lower():
                        dylib_dir = dylib_dir + ':' + resource_filename(epanet_toolkit, "libepanet/darwin-arm")
                    else:
                        dylib_dir = dylib_dir + ':' + resource_filename(epanet_toolkit, "libepanet/darwin-x64")
                    os.environ['DYLD_FALLBACK_LIBRARY_PATH'] = dylib_dir
            else:
                if os.name in ["nt", "dos"]:
                    libepanet = files(epanet_toolkit).joinpath("libepanet/windows-x64/epanet2.dll")
                    libmsx = files(epanet_toolkit).joinpath("libepanet/windows-x64/epanetmsx.dll")
                elif sys.platform in ["darwin"]:
                    if 'arm' in platform.platform().lower():
                        libepanet = files(epanet_toolkit).joinpath("libepanet/darwin-arm/libepanet2.dylib")
                        libmsx = files(epanet_toolkit).joinpath("libepanet/darwin-arm/libepanetmsx.dylib")
                    else:
                        libepanet = files(epanet_toolkit).joinpath("libepanet/darwin-x64/libepanet2.dylib")
                        libmsx = files(epanet_toolkit).joinpath("libepanet/darwin-x64/libepanetmsx.dylib")
                else:
                    libepanet = files(epanet_toolkit).joinpath("libepanet/linux-x64/libepanet2.so")
                    libmsx = files(epanet_toolkit).joinpath("libepanet/linux-x64/libepanetmsx.so")

                dylib_dir = os.environ.get('DYLD_FALLBACK_LIBRARY_PATH','')
                if dylib_dir != '':
                    if 'arm' in platform.platform().lower():
                        dylib_dir = dylib_dir + ':' + files(epanet_toolkit).joinpath("libepanet/darwin-arm")
                    else:
                        dylib_dir = dylib_dir + ':' + files(epanet_toolkit).joinpath("libepanet/darwin-x64")
                    os.environ['DYLD_FALLBACK_LIBRARY_PATH'] = dylib_dir
            if os.name in ["nt", "dos"]:
                self.ENlib = ctypes.windll.LoadLibrary(libmsx)
            else:
                self.ENlib = ctypes.cdll.LoadLibrary(libmsx)
        except:
            raise
        finally:
            self._project = None
        return

    def _error(self, *args):
        """Print the error text the corresponds to the error code returned"""
        if not self.errcode:
            return
        # errtxt = self.ENlib.ENgeterror(self.errcode)
        errtext =  MSX_ERROR_CODES.get(self.errcode, 'unknown error')
        if '%' in errtext and len(args) == 1:
            errtext % args
        if self.errcode >= 100:
            self.Errflag = True
            logger.error("EPANET error {} - {}".format(self.errcode, errtext))
            raise EpanetMsxException(self.errcode)
        return


    def ENopen(self, inpfile=None, rptfile=None, binfile=None):
        """
        Opens an EPANET input file and reads in network data

        Parameters
        ----------
        inpfile : str
            EPANET INP file (default to constructor value)
        rptfile : str
            Output file to create (default to constructor value)
        binfile : str
            Binary output file to create (default to constructor value)
        """
        if self.fileLoaded:
            self.ENclose()
        if self.fileLoaded:
            raise RuntimeError("File is loaded and cannot be closed")
        if inpfile is None:
            inpfile = self.inpfile
        if rptfile is None:
            rptfile = self.rptfile
        if binfile is None:
            binfile = self.binfile
        inpfile = inpfile.encode("latin-1")
        rptfile = rptfile.encode("latin-1")
        binfile = binfile.encode("latin-1")
        self.errcode = self.ENlib.MSXENopen(inpfile, rptfile, binfile)
        self._error()
        if self.errcode < 100:
            self.fileLoaded = True
        return

    def ENclose(self):
        """Frees all memory and files used by EPANET"""
        self.errcode = self.ENlib.MSXENclose()
        self._error()
        if self.errcode < 100:
            self.fileLoaded = False
        return

    # ----------running the simulation-----------------------------------------
    def MSXopen(self, msxfile):
        """Opens the MSX Toolkit to analyze a particular distribution system.

        Parameters
        ----------
        msxfile : str
            Name of the MSX input file
        """
        if msxfile is not None:
            msxfile = ctypes.c_char_p(msxfile.encode())
        ierr = self.ENlib.MSXopen(msxfile)
        if ierr != 0:
            raise EpanetMsxException(ierr, msxfile)

    def MSXclose(self):
        """Closes down the Toolkit system (including all files being processed)"""
        ierr = self.ENlib.MSXclose()
        if ierr != 0:
            raise EpanetMsxException(ierr)

    def MSXusehydfile(self, filename):
        """Uses the contents of the specified file as the current binary
        hydraulics file

        Parameters
        ----------
        filename : str
            Name of the hydraulics file to use
        """
        ierr = self.ENlib.MSXusehydfile(ctypes.c_char_p(filename.encode()))
        if ierr != 0:
            raise EpanetMsxException(ierr, filename)

    def MSXsolveH(self):
        """Runs a complete hydraulic simulation with results
        for all time periods written to the binary Hydraulics file."""
        ierr = self.ENlib.MSXsolveH()
        if ierr != 0:
            raise EpanetMsxException(ierr)

    def MSXinit(self, saveFlag=0):
        """Initializes the MSX system before solving for water quality results
        in step-wise fashion set saveFlag to 1 if water quality results should
        be saved to a scratch binary file, or to 0 is not saved to file"""
        saveFlag = int(saveFlag)
        ierr = self.ENlib.MSXinit(saveFlag)
        if ierr != 0:
            raise EpanetMsxException(ierr)

    def MSXsolveQ(self):
        """Solves for water quality over the entire simulation period and saves
        the results to an internal scratch file"""
        ierr = self.ENlib.MSXsolveQ()
        if ierr != 0:
            raise EpanetMsxException(ierr)

    def MSXstep(self):
        """Advances the water quality simulation one water quality time step.
        The time remaining in the overall simulation is returned as tleft, the
        current time as t."""
        t = ctypes.c_long()
        tleft = ctypes.c_long()
        ierr = self.ENlib.MSXstep(ctypes.byref(t), ctypes.byref(tleft))
        if ierr != 0:
            raise EpanetMsxException(ierr)
        out = [t.value, tleft.value]
        return out

    def MSXsaveoutfile(self, filename):
        """Saves water quality results computed for each node, link and
        reporting time period to a named binary file

        Parameters
        ----------
        filename : str
            Save a binary results file
        """
        ierr = self.ENlib.MSXsaveoutfile(ctypes.c_char_p(filename.encode()))
        if ierr != 0:
            raise EpanetMsxException(ierr)

    def MSXsavemsxfile(self, filename):
        """Saves the data associated with the current MSX project into a new
        MSX input file

        Parameters
        ----------
        filename : str
            Name of the MSX input file to create
        """
        ierr = self.ENlib.MSXsavemsxfile(ctypes.c_char_p(filename.encode()))
        if ierr != 0:
            raise EpanetMsxException(ierr, filename)

    def MSXreport(self):
        """Writes water quality simulations results as instructed by the MSX
        input file to a text file"""
        ierr = self.ENlib.MSXreport()
        if ierr != 0:
            raise EpanetMsxException(ierr)

    # ---------get parameters--------------------------------------------------
    def MSXgetindex(self, _type: Union[int, TkObjectType], name):
        """Gets the internal index of an MSX object given its name.

        Parameters
        ----------
        _type : int, str or ObjectType
            Type of object to get an index for
        name : str 
            Name of the object to get an index for

        Returns
        -------
        int
            Internal index
        
        Raises
        ------
        MSXKeyError
            If an invalid str is passed for _type
        MSXValueError
            If _type is not a valid MSX object type
        """
        try:
            _type = TkObjectType.get(_type)
        except KeyError:
            raise MSXKeyError(515, repr(_type))
        type_ind = int(_type)
        ind = ctypes.c_int()
        ierr = self.ENlib.MSXgetindex(type_ind, ctypes.c_char_p(name.encode()), ctypes.byref(ind))
        if ierr != 0:
            raise EpanetMsxException(ierr, repr(dict(_type=_type, name=name)))
        return ind.value

    def MSXgetIDlen(self, _type, index):
        """Get the number of characters in the ID name of an MSX object
        given its internal index number.

        Parameters
        ----------
        _type : int, str or ObjectType
            Type of object to get an index for
        index : int
            Index of the object to get the ID length for

        Returns
        -------
        int
            Length of the object ID
        """
        try:
            _type = TkObjectType.get(_type)
        except KeyError:
            raise MSXKeyError(515, repr(_type))
        type_ind = int(_type)
        len = ctypes.c_int()
        ierr = self.ENlib.MSXgetIDlen(type_ind, ctypes.c_int(index), ctypes.byref(len))
        if ierr != 0:
            raise EpanetMsxException(ierr, repr(dict(_type=_type, index=index)))
        return len.value

    def MSXgetID(self, _type, index):
        """Get the ID name of an object given its internal index number

        Parameters
        ----------
        _type : int, str or ObjectType
            Type of object to get an index for
        index : int
            Index of the object to get the ID for

        Returns
        -------
        str
            Object ID
        """
        try:
            _type = TkObjectType.get(_type)
        except KeyError:
            raise MSXKeyError(515, repr(_type))
        type_ind = int(_type)
        maxlen = 32
        id = ctypes.create_string_buffer(maxlen)
        ierr = self.ENlib.MSXgetID(type_ind, ctypes.c_int(index), ctypes.byref(id), ctypes.c_int(maxlen - 1))
        if ierr != 0:
            raise EpanetMsxException(ierr, repr(dict(_type=_type, index=index)))
        # the .decode() added my MF 6/3/21
        return id.value.decode()

    def MSXgetinitqual(self, _type, node_link_index, species_index):
        """Get the initial concentration of a particular chemical species
        assigned to a specific node or link of the pipe network

        Parameters
        ----------
        _type : str, int or ObjectType
            Type of object
        node_link_index : int
            Object index
        species_index : int
            Species index

        Returns
        -------
        float
            Initial quality value for that node or link

        Raises
        ------
        MSXKeyError
            Type passed in for ``_type`` is not valid
        MSXValueError
            Value for ``_type`` is not valid
        EpanetMsxException
            Any other error from the C-API
        """
        try:
            _type = TkObjectType.get(_type)
        except KeyError:
            raise MSXKeyError(515, repr(_type))
        if _type not in [TkObjectType.NODE, TkObjectType.LINK]:
            raise MSXValueError(515, repr(_type))
        type_ind = int(_type)
        iniqual = ctypes.c_double()
        ierr = self.ENlib.MSXgetinitqual(ctypes.c_int(type_ind), ctypes.c_int(node_link_index), ctypes.c_int(species_index), ctypes.byref(iniqual))
        if ierr != 0:
            raise EpanetMsxException(ierr, repr(dict(_type=_type, node_link_index=node_link_index, species_index=species_index)))
        return iniqual.value

    def MSXgetqual(self, _type, node_link_index, species_index):
        """Get a chemical species concentration at a given node or the
        average concentration along a link at the current simulation time step

        Parameters
        ----------
        _type : str, int or ObjectType
            Type of object
        node_link_index : int
            Object index
        species_index : int
            Species index

        Returns
        -------
        float
            Current quality value for that node or link

        Raises
        ------
        MSXKeyError
            Type passed in for ``_type`` is not valid
        MSXValueError
            Value for ``_type`` is not valid
        EpanetMsxException
            Any other error from the C-API
        """
        try:
            _type = TkObjectType.get(_type)
        except KeyError:
            raise MSXKeyError(515, repr(_type))
        if _type not in [TkObjectType.NODE, TkObjectType.LINK]:
            raise MSXValueError(515, repr(_type))
        type_ind = int(_type)
        qual = ctypes.c_double()
        ierr = self.ENlib.MSXgetqual(ctypes.c_int(type_ind), ctypes.c_int(node_link_index), ctypes.c_int(species_index), ctypes.byref(qual))
        if ierr != 0:
            raise EpanetMsxException(ierr, repr(dict(_type=_type, node_link_index=node_link_index, species_index=species_index)))
        return qual.value

    def MSXgetconstant(self, constant_index):
        """Get the value of a particular reaction constant

        Parameters
        ----------
        constant_index : int
            Index to the constant

        Returns
        -------
        float
            Value of the constant

        Raises
        ------
        EpanetMsxException
            Toolkit error occurred
        """
        const = ctypes.c_double()
        ierr = self.ENlib.MSXgetconstant(constant_index, ctypes.byref(const))
        if ierr != 0:
            raise EpanetMsxException(ierr, constant_index)
        return const.value

    def MSXgetparameter(self, _type, node_link_index, param_index):
        """Get the value of a particular reaction parameter for a given
        TANK or PIPE.

        Parameters
        ----------
        _type : int or str or Enum
            Get the type of the parameter
        node_link_index : int
            Link index
        param_index : int
            Parameter variable index

        Returns
        -------
        float
            Parameter value

        Raises
        ------
        MSXKeyError
            If there is no such _type
        MSXValueError
            If the _type is improper
        EpanetMsxException
            Any other error
        """
        try:
            _type = TkObjectType.get(_type)
        except KeyError:
            raise MSXKeyError(515, repr(_type))
        if _type not in [TkObjectType.NODE, TkObjectType.LINK]:
            raise MSXValueError(515, repr(_type))
        type_ind = int(_type)
        param = ctypes.c_double()
        ierr = self.ENlib.MSXgetparameter(ctypes.c_int(type_ind), ctypes.c_int(node_link_index), ctypes.c_int(param_index), ctypes.byref(param))
        if ierr != 0:
            raise EpanetMsxException(ierr, repr(dict(_type=_type, node_link_index=node_link_index, param_index=param_index)))
        return param.value

    def MSXgetsource(self, node_index, species_index):
        """Get information on any external source of a particular
        chemical species assigned to a specific node of the pipe network
        
        Parameters
        ----------
        node_index : int
            Node index
        species_index : int
            Species index
            
        Returns
        -------
        list
            [source type, level, and pattern] where level is the baseline 
            concentration (or mass flow rate) of the source and pattern the 
            index of the time pattern used to add variability to the source's 
            baseline level (0 if no pattern defined for the source)
        """
        level = ctypes.c_double()
        _type = ctypes.c_int()
        pat = ctypes.c_int()
        ierr = self.ENlib.MSXgetsource(ctypes.c_int(node_index), ctypes.c_int(species_index), ctypes.byref(_type), ctypes.byref(level), ctypes.byref(pat))
        if ierr != 0:
            raise EpanetMsxException(ierr, repr(dict(node_index=node_index, species_index=species_index)))
        src_out = [TkSourceType.get(_type.value), level.value, pat.value]
        return src_out

    def MSXgetpatternlen(self, pat):
        """Get the number of time periods within a SOURCE time pattern.

        Parameters
        ----------
        pat : int
            Pattern index

        Returns
        -------
        int
            Number of time periods in the pattern
        """
        len = ctypes.c_int()
        ierr = self.ENlib.MSXgetpatternlen(pat, ctypes.byref(len))
        if ierr != 0:
            raise EpanetMsxException(ierr)
        return len.value

    def MSXgetpatternvalue(self, pat, period):
        """Get the multiplier at a specific time period for a given
        SOURCE time pattern

        Parameters
        ----------
        pat : int
            Pattern index
        period : int
            1-indexed period of the pattern to retrieve

        Returns
        -------
            Multiplier
        """
        val = ctypes.c_double()
        ierr = self.ENlib.MSXgetpatternvalue(pat, period, ctypes.byref(val))
        if ierr != 0:
            raise EpanetMsxException(ierr)
        return val.value

    def MSXgetcount(self, _type):
        """Get the number of objects of a specified type.

        Parameters
        ----------
        _type : int or str or Enum
            Type of object to count

        Returns
        -------
        int
            Number of objects of specified type

        Raises
        ------
        MSXKeyError
            If the _type is invalid
        """
        try:
            _type = TkObjectType.get(_type)
        except KeyError:
            raise MSXKeyError(515, repr(_type))
        type_ind = int(_type)
        count = ctypes.c_int()
        ierr = self.ENlib.MSXgetcount(type_ind, ctypes.byref(count))
        if ierr != 0:
            raise EpanetMsxException(ierr)
        return count.value

    def MSXgetspecies(self, species_index):
        """Get the attributes of a chemical species given its internal
        index number.

        Parameters
        ----------
        species_index : int
            Species index to query (starting from 1 as listed in the MSX input
            file)

        Returns
        -------
        int, str, float, float
            Type, units, aTol, and rTol for the species
        """
        type_ind = ctypes.c_int()
        units = ctypes.create_string_buffer(15)
        aTol = ctypes.c_double()
        rTol = ctypes.c_double()
        ierr = self.ENlib.MSXgetspecies(species_index, ctypes.byref(type_ind), ctypes.byref(units), ctypes.byref(aTol), ctypes.byref(rTol))
        if ierr != 0:
            raise EpanetMsxException(ierr)
        spe_out = [type_ind.value, units.value, aTol.value, rTol.value]
        return spe_out

    def MSXgeterror(self, errcode, len=100):
        """Get the text for an error message given its error code

        Parameters
        ----------
        errcode : int
            Error code
        len : int, optional
            Length of the error message, by default 100 and minimum 80

        Returns
        -------
        str
            String decoded from the DLL

        Warning
        -------
        Getting string parameters in this way is not recommended, because it
        requires setting up string arrays that may or may not be the correct
        size. Use the wntr.epanet.msx.enums package to get error information.
        """
        errmsg = ctypes.create_string_buffer(len)
        self.ENlib.MSXgeterror(errcode, ctypes.byref(errmsg), len)
        return errmsg.value.decode()

    # --------------set parameters-----------------------------------

    def MSXsetconstant(self, ind, value):
        """Set a new value to a specific reaction constant

        Parameters
        ----------
        ind : int
            Index to the variable
        value : float
            Value to give the constant
        """
        ierr = self.ENlib.MSXsetconstant(ctypes.c_int(ind), ctypes.c_double(value))
        if ierr != 0:
            raise EpanetMsxException(ierr)

    def MSXsetparameter(self, _type, ind, param, value):
        """Set a value to a particular reaction parameter for a given TANK
        or PIPE

        Parameters
        ----------
        _type : int or str or enum
            Type of value to set
        ind : int
            Tank or pipe index
        param : int
            Parameter variable index
        value : float
            Value to be set

        Raises
        ------
        MSXKeyError
            If there is no such _type
        MSXValueError
            If the _type is invalid
        """
        try:
            _type = TkObjectType.get(_type)
        except KeyError:
            raise MSXKeyError(515, repr(_type))
        if _type not in [TkObjectType.NODE, TkObjectType.LINK]:
            raise MSXValueError(515, repr(_type))
        type_ind = int(_type)
        ierr = self.ENlib.MSXsetparameter(ctypes.c_int(type_ind), ctypes.c_int(ind), ctypes.c_int(param), ctypes.c_double(value))
        if ierr != 0:
            raise EpanetMsxException(ierr)

    def MSXsetinitqual(self, _type, ind, spe, value):
        """Set the initial concentration of a particular chemical species
        assigned to a specific node or link of the pipe network.

        Parameters
        ----------
        _type : int or str or enum
            Type of network element to set
        ind : int
            Index of the network element
        spe : int
            Index of the species
        value : float
            Initial quality value
        """
        try:
            _type = TkObjectType.get(_type)
        except KeyError:
            raise MSXKeyError(515, repr(_type))
        if _type not in [TkObjectType.NODE, TkObjectType.LINK]:
            raise MSXValueError(515, repr(_type))
        type_ind = int(_type)
        ierr = self.ENlib.MSXsetinitqual(ctypes.c_int(type_ind), ctypes.c_int(ind), ctypes.c_int(spe), ctypes.c_double(value))
        if ierr != 0:
            raise EpanetMsxException(ierr)

    def MSXsetsource(self, node, spe, _type, level, pat):
        """Set the attributes of an external source of a particular chemical
        species in a specific node of the pipe network

        Parameters
        ----------
        node : int
            Node index
        spe : int
            Species index
        _type : int or str or enum
            Type of source
        level : float
            Source quality value
        pat : int
            Pattern index
        """
        try:
            _type = TkSourceType.get(_type)
        except KeyError:
            raise MSXKeyError(515, repr(_type))
        type_ind = int(_type)
        ierr = self.ENlib.MSXsetsource(ctypes.c_int(node), ctypes.c_int(spe), ctypes.c_int(type_ind), ctypes.c_double(level), ctypes.c_int(pat))
        if ierr != 0:
            raise EpanetMsxException(ierr)

    def MSXsetpattern(self, pat, mult):
        """Set multipliers to a given MSX SOURCE time pattern

        Parameters
        ----------
        pat : int
            Pattern index
        mult : list-like
            Pattern multipliers
        """
        length = len(mult)
        cfactors_type = ctypes.c_double * length
        cfactors = cfactors_type()
        for i in range(length):
            cfactors[i] = float(mult[i])
        ierr = self.ENlib.MSXsetpattern(ctypes.c_int(pat), cfactors, ctypes.c_int(length))
        if ierr != 0:
            raise EpanetMsxException(ierr)

    def MSXsetpatternvalue(self, pat, period, value):
        """Set the multiplier factor for a specific period within a SOURCE time
        pattern.

        Parameters
        ----------
        pat : int
            Pattern index
        period : int
            1-indexed pattern time period index
        value : float
            Value to set at that time period
        """
        ierr = self.ENlib.MSXsetpatternvalue(ctypes.c_int(pat), ctypes.c_int(period), ctypes.c_double(value))
        if ierr != 0:
            raise EpanetMsxException(ierr)

    def MSXaddpattern(self, patternid):
        """Add a new, empty MSX source time pattern to an MSX project.

        Parameters
        ----------
        patternid : str
            Name of the new pattern
        """
        ierr = self.ENlib.MSXaddpattern(ctypes.c_char_p(patternid.encode()))
        if ierr != 0:
            raise EpanetMsxException(ierr)



"""
The wntr.morph.skel module contains functions to skeletonize water
network models.
"""
import logging
import copy
import itertools
import networkx as nx
    
from wntr.network.elements import Pipe, Junction
from wntr.sim.core import WNTRSimulator
from wntr.sim import EpanetSimulator

logger = logging.getLogger(__name__)

def skeletonize(wn, pipe_diameter_threshold, branch_trim=True, series_pipe_merge=True, 
                parallel_pipe_merge=True, max_cycles=None, use_epanet=True,
                pipes_to_exclude:list=[], junctions_to_exclude:list=[],
                return_map=False, return_copy=True):
    """
    Perform network skeletonization using branch trimming, series pipe merge, 
    and parallel pipe merge operations. Candidate pipes for removal is based 
    on a pipe diameter threshold.  
        
    Parameters
    -------------
    wn: wntr WaterNetworkModel
        Water network model
    pipe_diameter_threshold: float 
        Pipe diameter threshold. Pipes with diameter <= threshold are 
        candidates for removal
    branch_trim: bool, optional
        If True, include branch trimming in skeletonization
    series_pipe_merge: bool, optional
        If True, include series pipe merge in skeletonization
    parallel_pipe_merge: bool, optional
        If True, include parallel pipe merge in skeletonization
    max_cycles: int or None, optional
        Maximum number of cycles in the skeletonization process. 
        One cycle performs branch trimming for all candidate pipes, followed
        by series pipe merging for all candidate pipes, followed by parallel 
        pipe merging for all candidate pipes. If max_cycles is set to None, 
        skeletonization will run until the network can no longer be reduced.
    use_epanet: bool, optional
        If True, use the EpanetSimulator to compute headloss in pipes.  
        If False, use the WNTRSimulator to compute headloss in pipes.
    pipes_to_exclude: list, optional
        List of pipe names to exclude from skeletonization
    junctions_to_exclude: list, optional
        List of junction names to exclude from skeletonization
    return_map: bool, optional
        If True, return a skeletonization map. The map is a dictionary 
        that includes original nodes as keys and a list of skeletonized nodes 
        that were merged into each original node as values.
    return_copy: bool, optional
        If True, modify and return a copy of the WaterNetworkModel object.
        If False, modify and return the original WaterNetworkModel object.
        
    Returns
    --------
    wntr WaterNetworkModel
        Skeletonized water network model
    dictionary
        Skeletonization map (if return_map = True) which includes original 
        nodes as keys and a list of skeletonized nodes that were merged into 
        each original node as values.
    """
    if len(pipes_to_exclude) > 0:
        assert len(set(pipes_to_exclude) - set(wn.pipe_name_list)) == 0
    if len(junctions_to_exclude) > 0:
        assert len(set(junctions_to_exclude) - set(wn.junction_name_list)) == 0
    
    skel = _Skeletonize(wn, use_epanet, return_copy, pipes_to_exclude, junctions_to_exclude)
    
    skel.run(pipe_diameter_threshold, branch_trim, series_pipe_merge, 
             parallel_pipe_merge, max_cycles)
    
    if return_map:
        return skel.wn, skel.skeleton_map
    else:
        return skel.wn

		
class _Skeletonize(object):
    
    def __init__(self, wn, use_epanet, return_copy, pipes_to_exclude, junctions_to_exclude):
        
        if return_copy:
            # Get a copy of the WaterNetworkModel
            self.wn = copy.deepcopy(wn)
        else:
            self.wn = wn
        
        # Get the WaterNetworkModel graph
        G = self.wn.to_graph()
        G = G.to_undirected()
        self.G = G
        
        # Create a map of original nodes to skeletonized nodes
        skel_map = {}
        for node_name in self.wn.node_name_list:
            skel_map[node_name] = [node_name]
        self.skeleton_map = skel_map

        # Get a list of junction and pipe names that are associated with controls
        # Add them to junctions and pipes to exclude
        junc_with_controls = []
        pipe_with_controls = []
        for name, control in self.wn.controls():
            for req in control.requires():
                if isinstance(req, Junction):
                    junc_with_controls.append(req.name)
                elif isinstance(req, Pipe):
                    pipe_with_controls.append(req.name)
        self.junc_to_exclude = list(set(junc_with_controls))
        self.junc_to_exclude.extend(junctions_to_exclude)
        self.pipe_to_exclude = list(set(pipe_with_controls))
        self.pipe_to_exclude.extend(pipes_to_exclude)
        
        # Calculate pipe headloss using a single period EPANET simulation
        duration = self.wn.options.time.duration
        if use_epanet:
            sim = EpanetSimulator(self.wn)
        else:
            sim = WNTRSimulator(self.wn)
        self.wn.options.time.duration = 0
        results = sim.run_sim()
        head = results.node['head']
        headloss = {}
        for link_name, link in self.wn.links():
            headloss[link_name] = float(abs(head.loc[0,link.start_node_name] - 
                                            head.loc[0,link.end_node_name]))
        self.headloss = headloss
        self.wn.options.time.duration = duration
    
        self.num_branch_trim = 0
        self.num_series_merge = 0
        self.num_parallel_merge = 0
        
		
    def run(self, pipe_threshold, branch_trim=True, series_pipe_merge=True, 
                parallel_pipe_merge=True, max_cycles=None):
        """
        Run iterative branch trim, series pipe merge, and parallel pipe merge 
        operations based on a pipe diameter threshold.  
        """
        num_junctions = self.wn.num_junctions
        iteration = 0
        flag = True
        
        while flag:
            if branch_trim:
                self.branch_trim(pipe_threshold)
            if series_pipe_merge:
                self.series_pipe_merge(pipe_threshold)
            if parallel_pipe_merge:
                self.parallel_pipe_merge(pipe_threshold)
            
            iteration = iteration + 1
            
            if (max_cycles is not None) and (iteration > max_cycles):
                flag = False
            if num_junctions == self.wn.num_junctions:
                flag = False
            else:
                num_junctions = self.wn.num_junctions

        return self.wn, self.skeleton_map
    
	
    def branch_trim(self, pipe_threshold):
        """
        Run a single branch trim operation based on a pipe diameter threshold.
        Branch trimming removes dead-end pipes smaller than the pipe 
        diameter threshold and redistributes demands (and associated demand 
        patterns) to the neighboring junction.
        """
        for junc_name in self.wn.junction_name_list:
            if junc_name in self.junc_to_exclude:
                continue
            neighbors = list(nx.neighbors(self.G,junc_name))
            if len(neighbors) > 1:
                continue
            if len(neighbors) == 0:
                continue
            neigh_junc_name = neighbors[0] # only one neighbor
            nPipes = len(self.G.adj[junc_name][neigh_junc_name])
            if nPipes > 1:
                continue
            neigh_junc = self.wn.get_node(neigh_junc_name)
            if not (isinstance(neigh_junc, Junction)):
                continue
            pipe_name = list(self.G.adj[junc_name][neigh_junc_name].keys())[0] # only one pipe
            pipe = self.wn.get_link(pipe_name)
            if not ((isinstance(pipe, Pipe)) and \
                (pipe.diameter <= pipe_threshold) and \
                pipe_name not in self.pipe_to_exclude):
                continue
            
            logger.info('Branch trim: '+ str(junc_name) + str(neighbors))
            
            # Update skeleton map        
            self.skeleton_map[neigh_junc_name].extend(self.skeleton_map[junc_name])
            self.skeleton_map[junc_name] = []
            
            # Move demand
            junc = self.wn.get_node(junc_name)
            for demand in junc.demand_timeseries_list:
                neigh_junc.demand_timeseries_list.append(demand)
            junc.demand_timeseries_list.clear()

            # Remove node and links from wn and G
            self.wn.remove_link(pipe_name, force=True)
            self.wn.remove_node(junc_name, force=True)
            self.G.remove_node(junc_name)
                    
            self.num_branch_trim +=1
            
        return self.wn, self.skeleton_map
    
	
    def series_pipe_merge(self, pipe_threshold):
        """
        Run a single series pipe merge operation based on a pipe diameter 
        threshold.  This operation combines pipes in series if both pipes are 
        smaller than the pipe diameter threshold. The larger diameter pipe is 
        retained, demands (and associated demand patterns) are redistributed 
        to the nearest junction.
        """
        for junc_name in self.wn.junction_name_list:
            if junc_name in self.junc_to_exclude:
                continue
            neighbors = list(nx.neighbors(self.G,junc_name))
            if not (len(neighbors) == 2):
                continue
            neigh_junc_name0 = neighbors[0]
            neigh_junc_name1 = neighbors[1]
            neigh_junc0 = self.wn.get_node(neigh_junc_name0)
            neigh_junc1 = self.wn.get_node(neigh_junc_name1)
            if not ((isinstance(neigh_junc0, Junction)) or \
               (isinstance(neigh_junc1, Junction))):
                continue
            pipe_name0 = list(self.G.adj[junc_name][neigh_junc_name0].keys())
            pipe_name1 = list(self.G.adj[junc_name][neigh_junc_name1].keys())
            if (len(pipe_name0) > 1) or (len(pipe_name1) > 1):
                continue
            pipe_name0 = pipe_name0[0] # only one pipe
            pipe_name1 = pipe_name1[0] # only one pipe
            pipe0 = self.wn.get_link(pipe_name0)
            pipe1 = self.wn.get_link(pipe_name1)
            if not ((isinstance(pipe0, Pipe)) and \
                (isinstance(pipe1, Pipe)) and \
                ((pipe0.diameter <= pipe_threshold) and \
                (pipe1.diameter <= pipe_threshold)) and \
                pipe_name0 not in self.pipe_to_exclude and \
                pipe_name1 not in self.pipe_to_exclude):
                continue
            # Find closest neighbor junction
            if (isinstance(neigh_junc0, Junction)) and \
               (isinstance(neigh_junc1, Junction)):
                if pipe0.length < pipe1.length:
                    closest_junc = neigh_junc0
                else:
                    closest_junc = neigh_junc1
            elif (isinstance(neigh_junc0, Junction)):
                closest_junc = neigh_junc0
            elif (isinstance(neigh_junc1, Junction)):
                closest_junc = neigh_junc1
            else:
                continue
            
            logger.info('Series pipe merge: ' + str(junc_name) + str(neighbors))
                
            # Update skeleton map    
            self.skeleton_map[closest_junc.name].extend(self.skeleton_map[junc_name])
            self.skeleton_map[junc_name] = []
                
            # Move demand
            junc = self.wn.get_node(junc_name)
            for demand in junc.demand_timeseries_list:
                closest_junc.demand_timeseries_list.append(demand)
            junc.demand_timeseries_list.clear()

            # Remove node and links from wn and G
            self.wn.remove_link(pipe_name0, force=True)
            self.wn.remove_link(pipe_name1, force=True)
            self.wn.remove_node(junc_name, force=True)
            self.G.remove_node(junc_name)
            
            # Compute new pipe properties
            props = self._series_merge_properties(pipe0, pipe1)
            
            # Add new pipe to wn and G
            dominant_pipe = self._select_dominant_pipe(pipe0, pipe1)
            self.wn.add_pipe(dominant_pipe.name, 
                             start_node_name=neigh_junc_name0, 
                             end_node_name=neigh_junc_name1, 
                             length=props['length'], 
                             diameter=props['diameter'], 
                             roughness=props['roughness'], 
                             minor_loss=props['minorloss'],
                             initial_status=props['status']) 
            self.G.add_edge(neigh_junc_name0, 
                            neigh_junc_name1, 
                            dominant_pipe.name)
            
            self.num_series_merge +=1
            
        return self.wn, self.skeleton_map
        
		
    def parallel_pipe_merge(self, pipe_threshold):
        """
        Run a single parallel pipe merge operation based on a pipe diameter 
        threshold.  This operation combines pipes in parallel if both pipes are 
        smaller than the pipe diameter threshold. The larger diameter pipe is 
        retained.
        """
        
        for junc_name in self.wn.junction_name_list:
            if junc_name in self.junc_to_exclude:
                continue
            neighbors = nx.neighbors(self.G,junc_name)
            for neighbor in [n for n in neighbors]:
                parallel_pipe_names = list(self.G.adj[junc_name][neighbor].keys())
                if len(parallel_pipe_names) == 1:
                    continue
                for (pipe_name0, pipe_name1) in itertools.combinations(parallel_pipe_names, 2):
                    try:
                        pipe0 = self.wn.get_link(pipe_name0)
                        pipe1 = self.wn.get_link(pipe_name1)
                    except:
                        continue # one of the pipes removed in previous loop
                    if not ((isinstance(pipe0, Pipe)) and \
                       (isinstance(pipe1, Pipe)) and \
                        ((pipe0.diameter <= pipe_threshold) and \
                        (pipe1.diameter <= pipe_threshold)) and \
                        pipe_name0 not in self.pipe_to_exclude and \
                        pipe_name1 not in self.pipe_to_exclude):
                        continue
                    
                    logger.info('Parallel pipe merge: '+ str(junc_name) + str((pipe_name0, pipe_name1)))

                    # Remove links from wn and G   
                    self.wn.remove_link(pipe_name0, force=True)
                    self.wn.remove_link(pipe_name1, force=True)
                    self.G.remove_edge(neighbor, junc_name, pipe_name0) 
                    self.G.remove_edge(junc_name, neighbor, pipe_name1)
            
                    # Compute new pipe properties
                    props = self._parallel_merge_properties(pipe0, pipe1)

                    # Add a new pipe to wn and G
                    dominant_pipe = self._select_dominant_pipe(pipe0, pipe1)
                    self.wn.add_pipe(dominant_pipe.name, 
                                     start_node_name=dominant_pipe.start_node_name, 
                                     end_node_name=dominant_pipe.end_node_name,
                                     length=props['length'], 
                                     diameter=props['diameter'], 
                                     roughness=props['roughness'], 
                                     minor_loss=props['minorloss'],
                                     initial_status=props['status']) 
                    self.G.add_edge(dominant_pipe.start_node_name, 
                                    dominant_pipe.end_node_name, 
                                    dominant_pipe.name)
                     
                    self.num_parallel_merge +=1
                    
        return self.wn, self.skeleton_map
    
	
    def _select_dominant_pipe(self, pipe0, pipe1):
	
        # Dominant pipe = larger diameter
        if pipe0.diameter >= pipe1.diameter:
            dominant_pipe = pipe0
        else:
            dominant_pipe = pipe1
            
        return dominant_pipe

		
    def _series_merge_properties(self, pipe0, pipe1):
        
        props = {}
        dominant_pipe = self._select_dominant_pipe(pipe0, pipe1)
            
        props['length'] = pipe0.length + pipe1.length
        props['diameter'] = dominant_pipe.diameter
        props['minorloss'] = dominant_pipe.minor_loss
        props['status'] = dominant_pipe.status
        
        props['roughness'] = (props['length']/(props['diameter']**4.87))**0.54 * \
            ((pipe0.length/((pipe0.diameter**4.87)*(pipe0.roughness**1.85))) + \
             (pipe1.length/((pipe1.diameter**4.87)*(pipe1.roughness**1.85))))**-0.54
        
        return props
         
		 
    def _parallel_merge_properties(self, pipe0, pipe1):
        
        props = {}
        dominant_pipe = self._select_dominant_pipe(pipe0, pipe1)
            
        props['length'] = dominant_pipe.length
        props['diameter'] = dominant_pipe.diameter
        props['minorloss'] = dominant_pipe.minor_loss
        props['status'] = dominant_pipe.status
        
        props['roughness'] = ((props['length']**0.54)/(props['diameter']**2.63)) * \
            ((pipe0.roughness*(pipe0.diameter**2.63))/(pipe0.length**0.54) + \
             (pipe1.roughness*(pipe1.diameter**2.63))/(pipe1.length**0.54))
        
        return props




import unittest
from os.path import abspath, dirname, join
#import matplotlib.pylab as plt
import wntr

testdir = dirname(abspath(str(__file__)))
datadir = join(testdir, "..", "..", "examples", "networks")

class TestSimulationResults(unittest.TestCase):
    
    @classmethod
    def setUpClass(self):
        inp_file = join(datadir, "Net2.inp")
        self.wn = wntr.network.WaterNetworkModel(inp_file)
        sim = wntr.sim.EpanetSimulator(self.wn)
        sim.run_sim()
        
        binfile = wntr.epanet.io.BinFile()
        results_GPM = binfile.read('temp.bin', False, False, False)
        # time index needs to be converted to hours
        for key in results_GPM.node.keys():
            results_GPM.node[key].index = results_GPM.node[key].index/3600
        for key in results_GPM.link.keys():
            results_GPM.link[key].index = results_GPM.link[key].index/3600

        self.results_GPM = results_GPM
        
        self.tol = {}
        self.tol['node'] = {}
        self.tol['node']['demand'] = 1e-3
        self.tol['node']['head'] = 1e-3
        self.tol['node']['pressure'] = 1e-3
        self.tol['node']['quality'] = 1e-3
        self.tol['link'] = {}
        self.tol['link']['flowrate'] = 1e-2
        self.tol['link']['velocity'] = 1e-3
        self.tol['link']['headloss'] = 1e-3
        self.tol['link']['status'] = None # status is not checked.
        self.tol['link']['setting'] = 1e-5
        self.tol['link']['friction_factor'] = 1e-5
        self.tol['link']['quality'] = 1e-5
        self.tol['link']['reaction_rate'] = 1e-3

    @classmethod
    def tearDownClass(self):
        pass
 


if __name__ == "__main__":
    unittest.main()




# coding: utf-8
"""
The wntr.epanet.msx package provides EPANET-MSX compatibility functions for 
WNTR.

The following environment variable must be set, or the command `set_msx_path` 
must be run prior to trying to instantiate the EPANET-MSX toolkit.

.. envvar:: WNTR_PATH_TO_EPANETMSX

    The full path to the directory where EPANET-MSX has been installed. 
    Specifically, the directory should contain both toolkit files, epanet2.dll 
    and epanetmsx.dll (or the appropriate equivalent files for your system 
    architecture).
"""

import os as _os

def set_msx_path(path):
    if not _os.path.isdir(path):
        raise FileNotFoundError('Directory not found, {}'.format(path))
    _os.environ['WNTR_PATH_TO_EPANETMSX'] = path

from .io import MsxFile, MsxBinFile
from .toolkit import MSXepanet




import math
import unittest
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
import wntr
from wntr.sim.models.utils import ModelUpdater

testdir = dirname(abspath(str(__file__)))
test_data_dir = join(testdir, "data_for_testing")


def compare_floats(a, b, tol=1e-5, rel_tol=1e-3):
    if abs(a) >= 1e-8:
        if abs(a - b) / abs(a) * 100 > rel_tol:
            return False
        return True
    if abs(a - b) > tol:
        return False
    return True


def approximate_derivative(con, var, h):
    orig_value = var.value
    var_vals = [orig_value - 2 * h, orig_value - h, orig_value + h, orig_value + 2 * h]
    f_vals = []
    for i in var_vals:
        var.value = i
        f_vals.append(con.evaluate())
    coeffs = [1, -8, 8, -1]
    var.value = orig_value
    return sum(coeffs[i] * f_vals[i] for i in range(len(f_vals))) / (12 * h)


def abs_hw(pipe, f, m):
    f = abs(f)
    if f >= m.hw_q2:
        r = (
            m.hw_k
            * pipe.roughness ** (-1.852)
            * pipe.diameter ** (-4.871)
            * pipe.length
            * (f ** 1.852)
        )
    elif f >= m.hw_q1:
        r = (
            m.hw_k
            * pipe.roughness ** (-1.852)
            * pipe.diameter ** (-4.871)
            * pipe.length
            * (m.hw_a * f ** 3 + m.hw_b * f ** 2 + m.hw_c * f + m.hw_d)
        )
    else:
        r = (
            m.hw_k
            * pipe.roughness ** (-1.852)
            * pipe.diameter ** (-4.871)
            * pipe.length
            * m.hw_m
            * f
        )
    return r


def abs_minor_loss(pipe, f):
    f = abs(f)
    return pipe.minor_loss / 2.0 / 9.81 * (f / (math.pi / 4 * pipe.diameter ** 2)) ** 2


class TestHeadloss(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.wn = wn = wntr.network.WaterNetworkModel()
        wn.add_tank("t1", 0, 10, 0, 20, 15)
        wn.add_junction("j1", 0.01)
        wn.add_pipe("p1", "t1", "j1", minor_loss=10.0)
        wn.add_curve("curve1", "HEAD", [(0.0, 10.0), (0.05, 5.0), (0.1, 0.0)])
        wn.add_curve("curve2", "HEAD", [(0.0, 10.0), (0.03, 5.0), (0.1, 0.0)])
        wn.add_curve("curve3", "HEAD", [(0.0, 10.0), (0.07, 5.0), (0.1, 0.0)])

        # add a single point, 2-point, and a set of multi-point curves to test
        wn.add_curve("curve4", "HEAD", [(0.05, 5.0)])
        wn.add_curve("curve5", "HEAD", [(0.0, 10.0), (0.1, 0.0)])

        # multi_point_pump_curves = pump_curves_for_testing() # change to read in a csv file

        df = pd.read_csv(join(test_data_dir, "pump_practice_curves.csv"), skiprows=5)
        multi_point_pump_curves = []
        for i in range(11):
            multi_point_pump_curves.append(df[df["curve number"] == i].iloc[:, 1:3])

        for i, curve in enumerate(multi_point_pump_curves):
            curve_name = "curve{0:d}".format(i + 6)
            wn.add_curve(curve_name, "HEAD", curve.values)

        wn.add_pump("pump1", "t1", "j1", "HEAD", "curve1")
        wn.add_pump("pump2", "t1", "j1", "POWER", 50.0)
        cls.m = m = wntr.sim.aml.Model()
        cls.updater = updater = ModelUpdater()
        wntr.sim.models.constants.hazen_williams_constants(m)
        wntr.sim.models.param.hw_resistance_param.build(m, wn, updater)
        wntr.sim.models.param.minor_loss_param.build(m, wn, updater)
        wntr.sim.models.param.source_head_param(m, wn)
        wntr.sim.models.var.flow_var(m, wn)
        wntr.sim.models.var.head_var(m, wn)
        wntr.sim.models.constraint.piecewise_hazen_williams_headloss_constraint.build(
            m, wn, updater
        )
        wntr.sim.models.constants.head_pump_constants(m)
        wntr.sim.models.constraint.head_pump_headloss_constraint.build(m, wn, updater)
        wntr.sim.models.param.pump_power_param.build(m, wn, updater)
        wntr.sim.models.constraint.power_pump_headloss_constraint.build(m, wn, updater)

    def test_HW_headloss(self):
        wn = self.wn
        m = self.m
        pipe = wn.get_link("p1")

        flows_to_test = [
            0,
            m.hw_q1 / 2.0,
            m.hw_q1,
            (m.hw_q1 + m.hw_q2) / 2.0,
            m.hw_q2,
            m.hw_q2 + 0.1,
        ]
        _flows_to_test = [-i for i in flows_to_test]
        flows_to_test.extend(_flows_to_test)
        status_to_test = [1, 0]

        t1_head = 10
        j1_head = 0

        m.source_head["t1"].value = t1_head
        m.head["j1"].value = j1_head

        for status in status_to_test:
            pipe._user_status = status
            self.updater.update(m, wn, pipe, "status")
            for f in flows_to_test:
                m.flow["p1"].value = f
                r1 = m.piecewise_hazen_williams_headloss["p1"].evaluate()
                if status == 1:
                    if f <= 0:
                        r2 = (
                            abs_hw(pipe, f, m)
                            + abs_minor_loss(pipe, f)
                            + t1_head
                            - j1_head
                        )
                    else:
                        r2 = (
                            -abs_hw(pipe, f, m)
                            - abs_minor_loss(pipe, f)
                            + t1_head
                            - j1_head
                        )
                else:
                    r2 = f
                self.assertAlmostEqual(r1, r2, 12)
                d1 = m.piecewise_hazen_williams_headloss["p1"].reverse_ad()[
                    m.flow["p1"]
                ]
                d2 = m.piecewise_hazen_williams_headloss["p1"].reverse_ad()[
                    m.flow["p1"]
                ]
                d3 = approximate_derivative(
                    m.piecewise_hazen_williams_headloss["p1"], m.flow["p1"], 1e-6
                )
                rel_diff = abs(d1 - d2) / abs(d1) * 100
                self.assertLess(rel_diff, 0.1)
                rel_diff = abs(d1 - d3) / abs(d3) * 100
                self.assertLess(rel_diff, 0.1)
                self.assertAlmostEqual(d1, d2, 3)
                self.assertAlmostEqual(d1, d3, 3)

    def test_head_pump_headloss(self):
        wn = self.wn
        m = self.m
        pump = wn.get_link("pump1")
        curves_to_test = []
        for curve_name, curve in wn.curves.items():
            if curve.curve_type == "HEAD":
                curves_to_test.append(curve)

        flows_to_test = [
            m.pump_q1 - 0.1,
            m.pump_q1,
            (m.pump_q1 + m.pump_q2) / 2,
            m.pump_q2,
            m.pump_q2 + 0.1,
        ]
        status_to_test = [1, 0]

        t1_head = 5
        j1_head = 10

        m.source_head["t1"].value = t1_head
        m.head["j1"].value = j1_head

        for curve in curves_to_test:
            pump.pump_curve_name = curve.name
            self.updater.update(m, wn, pump, "pump_curve_name")
            A, B, C = pump.get_head_curve_coefficients()
            for status in status_to_test:
                pump._user_status = status
                self.updater.update(m, wn, pump, "status")
                for f in flows_to_test:
                    m.flow["pump1"].value = f
                    r1 = m.head_pump_headloss["pump1"].evaluate()
                    if status == 1:
                        if C <= 1:
                            (
                                a,
                                b,
                                c,
                                d,
                            ) = wntr.sim.models.constraint.get_pump_poly_coefficients(
                                A, B, C, m
                            )
                            if f <= m.pump_q1:
                                r2 = m.pump_slope * f + A - j1_head + t1_head
                            elif f <= m.pump_q2:
                                r2 = (
                                    a * f ** 3
                                    + b * f ** 2
                                    + c * f
                                    + d
                                    - j1_head
                                    + t1_head
                                )
                            else:
                                r2 = A - B * f ** C - j1_head + t1_head
                        else:
                            (
                                q_bar,
                                h_bar,
                            ) = wntr.sim.models.constraint.get_pump_line_params(
                                A, B, C, m
                            )
                            if f <= q_bar:
                                r2 = (
                                    m.pump_slope * (f - q_bar)
                                    + h_bar
                                    - j1_head
                                    + t1_head
                                )
                            else:
                                r2 = A - B * f ** C - j1_head + t1_head
                    else:
                        r2 = f
                    self.assertTrue(compare_floats(r1, r2, 1e-12, 1e-12))
                    if f < m.pump_q1 or f > m.pump_q2:
                        d1 = m.head_pump_headloss["pump1"].reverse_ad()[m.flow["pump1"]]
                        d2 = m.head_pump_headloss["pump1"].reverse_ad()[m.flow["pump1"]]
                        d3 = approximate_derivative(
                            m.head_pump_headloss["pump1"], m.flow["pump1"], 1e-6
                        )
                        self.assertTrue(compare_floats(d1, d2, 1e-12, 1e-12))
                        self.assertTrue(compare_floats(d1, d3, 1e-5, 1e-3))

    def test_power_pump_headloss(self):
        wn = self.wn
        m = self.m
        pump = wn.get_link("pump2")

        flows_to_test = np.linspace(-1.0, 1.0, 10, True)
        status_to_test = [1, 0]

        t1_head = 5
        j1_head = 10

        m.source_head["t1"].value = t1_head
        m.head["j1"].value = j1_head

        for status in status_to_test:
            pump._user_status = status
            self.updater.update(m, wn, pump, "status")
            for f in flows_to_test:
                ff = float(f)  # expr.m does not bind to
                # numpy.float64 correctly
                m.flow["pump2"].value = ff

                r1 = m.power_pump_headloss["pump2"].evaluate()
                if status == 1:
                    r2 = pump.power + (t1_head - j1_head) * ff * 9.81 * 1000.0
                else:
                    r2 = ff
                self.assertTrue(compare_floats(r1, r2, 1e-12, 1e-12))
                d1 = m.power_pump_headloss["pump2"].reverse_ad()[m.flow["pump2"]]
                d2 = m.power_pump_headloss["pump2"].reverse_ad()[m.flow["pump2"]]
                d3 = approximate_derivative(
                    m.power_pump_headloss["pump2"], m.flow["pump2"], 1e-6
                )
                self.assertTrue(compare_floats(d1, d2, 1e-12, 1e-12))
                self.assertTrue(compare_floats(d1, d3, 1e-8, 1e-6))


class TestPDD(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.wn = wn = wntr.network.WaterNetworkModel()
        wn.add_junction("j1", 0.01)

    def test_pdd(self):
        wn = self.wn
        # Changed to handle the specific heads_to_test range, which is bad
        node = wn.get_node("j1")
        node.required_pressure = 20.0
        node.minimum_pressure = 0.0

        m = wntr.sim.aml.Model()
        updater = ModelUpdater()
        wntr.sim.models.constants.pdd_constants(m)
        wntr.sim.models.param.expected_demand_param(m, wn)
        wntr.sim.models.param.elevation_param.build(m, wn, updater)
        wntr.sim.models.param.pdd_poly_coeffs_param.build(m, wn, updater)
        wntr.sim.models.param.pmin_param.build(m, wn, updater)
        wntr.sim.models.param.pnom_param.build(m, wn, updater)
        wntr.sim.models.var.head_var(m, wn)
        wntr.sim.models.var.demand_var(m, wn)
        wntr.sim.models.constraint.pdd_constraint.build(m, wn, updater)

        node = wn.get_node("j1")

        pmin = node.minimum_pressure
        pnom = node.required_pressure
        h0 = node.elevation + pmin
        h1 = node.elevation + pnom
        delta = m.pdd_smoothing_delta
        heads_to_test = [
            h0 - 1,
            h0,
            h0 + delta / 2,
            h0 + delta,
            h0 + (pmin + pnom) / 2,
            h1 - delta,
            h1 - delta / 2,
            h1,
            h1 + 1,
        ]
        d_expected = node.demand_timeseries_list.at(
            wn.sim_time, multiplier=wn.options.hydraulic.demand_multiplier
        )

        d = 0.01

        m.demand["j1"].value = d
        a1 = m.pdd_poly1_coeffs_a["j1"].value
        b1 = m.pdd_poly1_coeffs_b["j1"].value
        c1 = m.pdd_poly1_coeffs_c["j1"].value
        d1 = m.pdd_poly1_coeffs_d["j1"].value
        a2 = m.pdd_poly2_coeffs_a["j1"].value
        b2 = m.pdd_poly2_coeffs_b["j1"].value
        c2 = m.pdd_poly2_coeffs_c["j1"].value
        d2 = m.pdd_poly2_coeffs_d["j1"].value
        delta = m.pdd_smoothing_delta
        slope = m.pdd_slope

        for h in heads_to_test:
            m.head["j1"].value = h
            r1 = m.pdd["j1"].evaluate()
            p = h - node.elevation
            if p <= pmin:
                r2 = d - d_expected * slope * (p - pmin)
            elif p <= pmin + delta:
                r2 = d - d_expected * (a1 * p ** 3 + b1 * p ** 2 + c1 * p + d1)
            elif p <= pnom - delta:
                r2 = d - d_expected * ((p - pmin) / (pnom - pmin)) ** 0.5
            elif p <= pnom:
                r2 = d - d_expected * (a2 * p ** 3 + b2 * p ** 2 + c2 * p + d2)
            else:
                r2 = d - d_expected * (slope * (p - pnom) + 1.0)
            # print(p, r1, r2, abs(r1 - r2))
            self.assertAlmostEqual(r1, r2, 12)
            der1 = m.pdd["j1"].reverse_ad()[m.head["j1"]]
            der2 = m.pdd["j1"].reverse_ad()[m.head["j1"]]
            der3 = approximate_derivative(m.pdd["j1"], m.head["j1"], 1e-6)
            self.assertAlmostEqual(der1, der2, 7)
            self.assertAlmostEqual(der1, der3, 6)


if __name__ == "__main__":
    unittest.main()
